import collections

import numpy as np

import treams.lattice as la


class Lattice:
    """Lattice definition.

    The lattice can be one-, two-, or three-dimensional. If it is not three-dimensional
    it is required to be embedded into a lower dimensional subspace that is aligned with
    the Cartesian axes. The default alignment for one-dimensional lattices is along the
    z-axis and for two-dimensional lattices it is in the x-y-plane.

    For a one-dimensional lattice the definition consists of simply the value of the
    lattice pitch. Higher-dimensional lattices are defined by (2, 2)- or (3, 3)-arrays.
    If the arrays are diagonal it is sufficient to specify the (2,)- or (3,)-array
    diagonals. Alternatively, if another instance of a Lattice is given, the defined
    alignment is extracted, which can be used to separate a lower-dimensional
    sublattice.

    Lattices are immutable objects.

    Example:
        >>> Lattice([1, 2])
        Lattice([[1. 0.]
                 [0. 2.]], alignment='xy')
        >>> Lattice(_, 'x')
        Lattice(1.0, alignment='x')

    Args:
        arr (float, array, Lattice): Lattice definition. Each row corresponds to one
            lattice vector, each column to the axis defined in `alignment`.
        alignment (str, optional): Alignment of the lattice. Possible values are 'x',
            'y', 'z', 'xy', 'yz', 'zx', and 'xyz'.
    """

    _allowed_alignments = {
        "x": "y",
        "y": "z",
        "z": "x",
        "xy": "yz",
        "yz": "zx",
        "zx": "xy",
        "xyz": "xyz",
    }

    def __init__(self, arr, alignment=None):
        """Initialization."""
        if isinstance(arr, Lattice):
            self._alignment = arr.alignment if alignment is None else alignment
            self._lattice = arr._sublattice(self._alignment)[...]
            self._reciprocal = (
                2 * np.pi / self[...] if self.dim == 1 else la.reciprocal(self[...])
            )
            return
        if arr is None:
            raise ValueError("Lattice cannot be 'None'")

        arr = np.array(arr, float)
        arr.flags.writeable = False
        alignments = {1: ("z", "x", "y"), 4: ("xy", "yz", "zx"), 9: ("xyz",)}
        if arr.ndim < 3 and arr.size == 1:
            self._lattice = np.squeeze(arr)
        elif arr.ndim == 1 and arr.size in (2, 3):
            self._lattice = np.diag(arr)
        elif arr.ndim == 2 and arr.shape[0] == arr.shape[1] and arr.shape[0] in (2, 3):
            self._lattice = arr
        else:
            raise ValueError(f"invalid shape '{arr.shape}'")

        alignment = (
            alignments[self._lattice.size][0] if alignment is None else alignment
        )
        if alignment not in alignments[self._lattice.size]:
            raise ValueError(f"invalid alignment '{alignment}'")
        self._alignment = alignment
        if self.volume == 0:
            raise ValueError("linearly dependent lattice vectors")
        self._reciprocal = (
            2 * np.pi / self[...] if self.dim == 1 else la.reciprocal(self[...])
        )

    @property
    def alignment(self):
        """The alignment of the lattice in three-dimensional space.

        For three-dimensional lattices it is always 'xyz' but lower-dimensional lattices
        have to be aligned with a subset of the axes.

        Returns:
            str
        """
        return self._alignment

    @classmethod
    def square(cls, pitch, alignment=None):
        """Create a two-dimensional square lattice.

        Args:
            pitch (float): Lattice constant.
            alignment (str, optional): Alignment of the two-dimensional lattice in the
                three-dimensional space. Defaults to 'xy'.

        Returns:
            Lattice
        """
        return cls(2 * [pitch], alignment)

    @classmethod
    def cubic(cls, pitch, alignment=None):
        """Create a three-dimensional cubic lattice.

        Args:
            pitch (float): Lattice constant.
            alignment (str, optional): Alignment of the lattice. Defaults to 'xyz'.

        Returns:
            Lattice
        """
        return cls(3 * [pitch], alignment)

    @classmethod
    def rectangular(cls, x, y, alignment=None):
        """Create a two-dimensional rectangular lattice.

        Args:
            x (float): Lattice constant along the first dimension. For the default
                alignment this corresponds to the x-axis.
            y (float): Lattice constant along the second dimension. For the default
                alignment this corresponds to the y-axis.
            alignment (str, optional): Alignment of the two-dimensional lattice in the
                three-dimensional space. Defaults to 'xy'.

        Returns:
            Lattice
        """
        return cls([x, y], alignment)

    @classmethod
    def orthorhombic(cls, x, y, z, alignment=None):
        """Create a three-dimensional orthorhombic lattice.

        Args:
            x (float): Lattice constant along the first dimension. For the default
                alignment this corresponds to the x-axis.
            y (float): Lattice constant along the second dimension. For the default
                alignment this corresponds to the y-axis.
            z (float): Lattice constant along the third dimension. For the default
                alignment this corresponds to the z-axis.
            alignment (str, optional): Alignment of the lattice. Defaults to 'xyz'.

        Returns:
            Lattice
        """
        return cls([x, y, z], alignment)

    @classmethod
    def hexagonal(cls, pitch, height=None, alignment=None):
        """Create a hexagonal lattice.

        The lattice is two-dimensional if no height is specified else it is
        three-dimensional

        Args:
            pitch (float): Lattice constant.
            height (float, optional): Separation along the third axis for a
                three-dimensional lattice.
            alignment (str, optional): Alignment of the two-dimensional lattice in the
                three-dimensional space. Defaults to either 'xy' or 'xyz' in the two-
                or three-dimensional case, respectively.

        Returns:
            Lattice
        """
        if height is None:
            return cls(
                np.array([[pitch, 0], [0.5 * pitch, np.sqrt(0.75) * pitch]]), alignment
            )
        return cls(
            np.array(
                [[pitch, 0, 0], [0.5 * pitch, np.sqrt(0.75) * pitch, 0], [0, 0, height]]
            ),
            alignment,
        )

    def __eq__(self, other):
        """Equality.

        Two lattices are considered equal when they have the same dimension, alignment,
        and lattice vectors.

        Args:
            other (Lattice): Lattice to compare with.

        Returns:
            bool
        """
        return other is not None and (
            self is other
            or (
                self.alignment == other.alignment
                and self.dim == other.dim
                and np.all(self[...] == other[...])
            )
        )

    @property
    def dim(self):
        """Dimension of the lattice.

        Returns:
            int
        """
        if self._lattice.ndim == 0:
            return 1
        return self._lattice.shape[0]

    @property
    def volume(self):
        """(Generalized) volume of the lattice.

        The value gives the lattice pitch, area, or volume depending on its dimension.
        The volume is signed.

        Returns:
            float
        """
        if self.dim == 1:
            return self._lattice
        return la.volume(self._lattice)

    @property
    def reciprocal(self):
        r"""Reciprocal lattice.

        The reciprocal lattice to a given lattice with dimension :math:`d` and lattice
        vectors :math:`\boldsymbol a_i` for :math:`i \in \{1, \dots, d\}` is defined by
        lattice vectors :math:`\boldsymbol b_j` with :math:`j \in \{1, \dots, d\}` such
        that :math:`\boldsymbol a_i \boldsymbol b_j = 2 \pi \delta_{ij}` is fulfilled.

        Returns:
            Lattice
        """
        return self._reciprocal

    def __getitem__(self, idx):
        """Index into the lattice.

        Indexing can be used to obtain entries from the lattice vector definitions or
        by using the Ellipsis or empty tuple to obtain the full array.

        Returns:
            float
        """
        return self._lattice[idx]

    def __str__(self):
        """String representation.

        Simply returns the lattice pitch or lattice vectors.

        Returns:
            str
        """
        return str(self._lattice)

    def __repr__(self):
        """Representation.

        The result can be used to recreate an equal instance.

        Returns:
            str
        """
        string = str(self._lattice).replace("\n", "\n        ")
        return f"Lattice({string}, alignment='{self.alignment}')"

    def _sublattice(self, key):
        """Get the sublattice defined by key.

        This function is called when one gives another instance of Lattice to the
        constructor.

        Args:
            key (str): Alignment of the sublattice to extract.

        Returns:
            Lattice
        """
        key = key.lower()
        if self.dim == 1:
            if key == self.alignment:
                return Lattice(self[...], key)
            raise ValueError(f"sublattice with key '{key}' not availale")
        idx = []
        for c in key:
            idx.append(self.alignment.find(c))
        if -1 in idx or key not in self._allowed_alignments:
            raise ValueError(f"sublattice with key '{key}' not availale")
        idx_opp = [i for i in range(self.dim) if i not in idx]
        mask = np.any(self[:, idx] != 0, axis=-1)
        if (
            (self[mask, :][:, idx_opp] != 0).any()
            or (self[np.logical_not(mask), :][:, idx] != 0).any()
            or len(idx) != sum(mask)
        ):
            raise ValueError("cannot determine sublattice")
        return Lattice(self[mask, :][:, idx], key)

    def permute(self, n=1):
        """Permute the lattice orientation.

        Get a new lattice with the alignment permuted.

        Examples:
            >>> Lattice.hexagonal(1, 2).permute()
            Lattice([[0.    1.    0.   ]
                     [0.    0.5   0.866]
                     [2.    0.    0.   ]], alignment='xyz')
            >>> Lattice.hexagonal(1).permute()
            Lattice([[1.    0.   ]
                     [0.5   0.866]], alignment='yz')

        Args:
            n (int, optional): Number of repeated permutations. Defaults to `1`.

        Returns:
            Lattice
        """
        if n != int(n):
            raise ValueError("'n' must be integer")
        n = n % 3
        lattice = self[...]
        alignment = self.alignment
        if self.dim == 3:
            while n > 0:
                lattice = lattice[:, [2, 0, 1]]
                n -= 1
        else:
            while n > 0:
                alignment = self._allowed_alignments[alignment]
                n -= 1
        return Lattice(lattice, alignment)

    def __bool__(self):
        """Lattice instances always equate to True.

        Returns:
            bool
        """
        return True

    def __or__(self, other):
        """Merge two lattices.

        Two lattices are combined into one if possible.

        Example:
            >>> Lattice(1, 'x') | Lattice(2)
            Lattice([[2. 0.]
                     [0. 1.]], alignment='zx')

        Args:
            other (Lattice): Lattice to merge.

        Returns:
            Lattice
        """
        if other is None or self == other:
            return Lattice(self)

        alignment = list({c for lat in (self, other) for c in lat.alignment})
        alignment = "".join(sorted(alignment))
        if alignment == "xz":
            alignment = "zx"

        ndim = len(alignment)

        if ndim == 2:
            if self.dim == 1 == other.dim:
                if alignment.find(self.alignment) == 0:
                    return Lattice([self[...], other[...]], alignment)
                return Lattice([other[...], self[...]], alignment)
            if self.dim == 2 and Lattice(self, other.alignment) == other:
                return Lattice(self)
            if other.dim == 2 and Lattice(other, self.alignment) == self:
                return Lattice(other)
        elif ndim == 3:
            if self.dim == 3 and Lattice(self, other.alignment) == other:
                return Lattice(self)
            if other.dim == 3 and Lattice(other, self.alignment) == self:
                return Lattice(other)
            if self.dim == 2 == other.dim:
                arr = np.zeros(3)
                for i, c in enumerate("xyz"):
                    if c in self.alignment:
                        la0 = Lattice(self, c)
                    else:
                        la0 = None
                    if c in other.alignment:
                        la1 = Lattice(other, c)
                    else:
                        la1 = None
                    if None not in (la0, la1) and la0 != la1:
                        raise ValueError("cannot combine lattices")
                    arr[i] = la0[...] if la0 is not None else la1[...]
                return Lattice(arr)
            arr = np.zeros((3, 3))
            la0, la1 = (self, other) if self.dim == 1 else (other, self)
            if la0.alignment == "x":
                arr[0, 0] = la0[...]
                arr[1:, 1:] = la1[...]
                return Lattice(arr)
            if la0.alignment == "y":
                arr[1, 1] = la0[...]
                arr[[[2], [0]], [2, 0]] = la1[...]
                return Lattice(arr)
            if la0.alignment == "z":
                arr[2, 2] = la0[...]
                arr[:2, :2] = la1[...]
                return Lattice(arr)
        raise ValueError("cannot combine lattices")

    def __and__(self, other):
        """Intersect two lattices.

        The intersection of two lattices is taken if possible.

        Example:
            >>> Lattice([1, 2]) & Lattice([2, 3], 'yz')
            Lattice(2.0, alignment='y')

        Args:
            other (Lattice): Lattice to intersect.

        Returns:
            Lattice
        """
        if other is None:
            return None
        if self == other:
            return Lattice(self)

        alignment = list({c for c in self.alignment if c in other.alignment})
        if len(alignment) == 0:
            raise ValueError("cannot combine lattices")
        alignment = "".join(sorted(alignment))
        if alignment == "xz":
            alignment = "zx"

        a, b = Lattice(self, alignment), Lattice(other, alignment)
        if a == b:
            return a
        raise ValueError("cannot combine lattices")

    def __le__(self, other):
        """Test if one lattice includes another.

        Example:
            >>> Lattice(3) <= Lattice([1, 2, 3])
            True

        Args:
            other (Lattice): Lattice to compare with.

        Returns:
            bool
        """
        try:
            lat = Lattice(other, self.alignment)
        except ValueError:
            return False
        return lat == self

    def isdisjoint(self, other):
        """Test if lattices are disjoint.

        Lattices are considered disjoint if their alignments are disjoint.

        Example:
            >>> Lattice([1, 2, 3]).isdisjoint(Lattice(1))
            False

        Args:
            other (Lattice): Lattice to compare with.

        Returns:
            bool
        """
        for c in other.alignment:
            if c in self.alignment:
                return False
        return True


class WaveVector(collections.abc.Sequence):
    def __init__(self, seq=(), alignment=None):
        try:
            length = len(seq)
        except TypeError:
            length = 1
            seq = [seq]
        if length == 3:
            self._vec = tuple(seq)
        elif length == 2:
            if alignment in ("xy", None):
                self._vec = (seq[0], seq[1], np.nan)
            elif alignment == "yz":
                self._vec = (np.nan, seq[0], seq[1])
            elif alignment == "zx":
                self._vec = (seq[1], np.nan, seq[0])
            else:
                raise ValueError(f"invalid alignment: {alignment}")
        elif length == 1:
            if alignment in ("z", None):
                self._vec = (np.nan, np.nan, seq[0])
            elif alignment == "y":
                self._vec = (np.nan, seq[0], np.nan)
            elif alignment == "x":
                self._vec = (seq[0], np.nan, np.nan)
            else:
                raise ValueError(f"invalid alignment: {alignment}")
        elif length == 0:
            self._vec = (np.nan,) * 3
        else:
            raise ValueError("invalid sequence")

    def __str__(self):
        return str(self._vec)

    def __repr__(self):
        return self.__class__.__name__ + str(self._vec)

    def __eq__(self, other):
        other = WaveVector(other)
        for a, b in zip(self, other):
            if a != b and not (np.isnan(a) and np.isnan(b)):
                return False
        return True

    def __len__(self):
        return 3

    def __getitem__(self, key):
        return self._vec[key]

    def __or__(self, other):
        other = WaveVector(other)
        seq = ()
        for a, b in zip(self, other):
            isnan = np.isnan(a) or np.isnan(b)
            if a != b and not isnan:
                raise ValueError("non-matching WaveVector")
            seq += (np.nan,) if isnan else (a,)
        return WaveVector(seq)

    def __and__(self, other):
        other = WaveVector(other)
        seq = ()
        for a, b in zip(self, other):
            if a != b and not (np.isnan(a) or np.isnan(b)):
                raise ValueError("non-matching WaveVector")
            seq += (b,) if np.isnan(a) else (a,)
        return WaveVector(seq)

    def permute(self, n=1):
        x, y, z = self
        n = n % 3
        for _ in range(n):
            x, y, z = z, x, y
        return WaveVector((x, y, z))

    def __le__(self, other):
        other = WaveVector(other)
        for a, b in zip(self, other):
            if a != b and not np.isnan(b):
                return False
        return True

    def isdisjoint(self, other):
        other = WaveVector(other)
        for a, b in zip(self, other):
            if not (np.isnan(a) or np.isnan(b)):
                return False
        return True



"""Miscellaneous functions.

.. autosummary::
   :toctree:

   basischange
   firstbrillouin1d
   firstbrillouin2d
   firstbrillouin3d
   pickmodes
   refractive_index
   wave_vec_z
"""

import numpy as np

import treams.lattice as la


def refractive_index(epsilon=1, mu=1, kappa=0):
    r"""Refractive index of a (chiral) medium.

    The refractive indeces in a chiral medium :math:`\sqrt{\epsilon\mu} \mp \kappa` are
    returned with the negative helicity result first.

    Args:
        epsilon (float or complex, array_like, optional): Relative permittivity,
            defaults to 1.
        mu (float or complex, array_like, optional): Relative permeability,
            defaults to 1.
        kappa (float or complex, array_like, optional): Chirality parameter,
            defaults to 0.

    Returns:
        float or complex, (2,)-array

    """
    epsilon = np.array(epsilon)
    n = np.sqrt(epsilon * mu)
    res = np.stack((n - kappa, n + kappa), axis=-1)
    res[np.imag(res) < 0] *= -1
    return res


def basischange(out, in_=None):
    """Coefficients for the basis change between helicity and parity modes.

    Args:
        out (3- or 4-tuple of (M,)-arrays): Output modes, the last array is taken as
            polarization.
        in_ (3- or 4-tuple of (N,)-arrays, optional): Input modes, if none are given,
            equal to the output modes

    Returns:
        float, ((M, N)-array
    """
    if in_ is None:
        in_ = out
    out = np.array([*zip(*out)])
    in_ = np.array([*zip(*in_)])
    res = np.zeros((out.shape[0], in_.shape[0]))
    out = out[:, None, :]
    sqhalf = np.sqrt(0.5)
    equal = (out[:, :, :-1] == in_[:, :-1]).all(axis=-1)
    minus = np.logical_and(out[:, :, -1] == in_[:, -1], in_[:, -1] == 0)
    res[equal] = sqhalf
    res[np.logical_and(equal, minus)] = -sqhalf
    return res


def pickmodes(out, in_):
    """Coefficients to pick modes.

    Args:
        out (3- or 4-tuple of (M,)-arrays): Output modes, the last array is taken as
            polarization.
        in_ (3- or 4-tuple of (N,)-arrays, optional): Input modes, if none are given,
            equal to the output modes

    Returns:
        float, ((M, N)-array
    """
    out = np.array([*zip(*out)])
    in_ = np.array([*zip(*in_)])
    return np.all(out[:, None, :] == in_, axis=-1)


# def wave_vec_zs(kpars, ks):
#     kpars = np.array(kpars)
#     ks = np.array(ks)
#     if ks.ndim == 0:
#         nmat = npol = 1
#     elif ks.ndim == 1:
#         npol = ks.shape[0]
#         nmat = 1
#     elif ks.ndim == 2:
#         nmat, npol = ks.shape
#     else:
#         raise ValueError("ks has invalid shape")
#     res = np.zeros((nmat, kpars.shape[0], npol), np.complex128)
#     for i, kpar in enumerate(kpars):
#         res[:, i, :] = wave_vec_z(kpar[0], kpar[1], ks)
#     return res


def wave_vec_z(kx, ky, k):
    r"""Z component of the wave vector with positive imaginary part.

    The result is :math:`k_z = \sqrt{k^2 - k_x^2 - k_y^2}` with
    :math:`\arg k_z \in \[ 0, \pi )`.

    Args:
        kx (float, array_like): X component of the wave vector
        ky (float, array_like): Y component of the wave vector
        k (float or complex, array_like): Wave number

    Returns:
        complex
    """
    kx = np.asarray(kx)
    ky = np.asarray(ky)
    k = np.asarray(k, complex)
    res = np.sqrt(k * k - kx * kx - ky * ky)
    if res.ndim == 0 and res.imag < 0:
        res = -res
    elif res.ndim > 0:
        res[np.imag(res) < 0] *= -1
    return res


def firstbrillouin1d(kpar, b):
    """Map wave vector to first Brillouin zone in 1D.

    Reduce the 1d wave vector (actually just a number) to the first Brillouin zone, i.e.
    the range `(-b/2, b/2]`

    Args:
        kpar (float64): (parallel) wave vector
        b (float64): reciprocal lattice vector

    Returns:
        float64
    """
    kpar -= b * np.round(kpar / b)
    if kpar > 0.5 * b:
        kpar -= b
    if kpar <= -0.5 * b:
        kpar += b
    return kpar


def firstbrillouin2d(kpar, b, n=2):
    """Map wave vector to first Brillouin zone in 2D.

    The reduction to the first Brillouin zone is first approximated roughly. From this
    approximated vector and its 8 neighbours, the shortest one is picked. As a
    sufficient approximation is not guaranteed (especially for extreme geometries),
    this process is iterated `n` times.

    Args:
        kpar (1d-array): parallel wave vector
        b (2d-array): reciprocal lattice vectors
        n (int): number of iterations

    Returns:
        (1d-array)
    """
    kparstart = kpar
    b1 = b[0, :]
    b2 = b[1, :]
    normsq1 = b1 @ b1
    normsq2 = b2 @ b2
    normsqp = (b1 + b2) @ (b1 + b2)
    normsqm = (b1 - b2) @ (b1 - b2)
    if (
        normsqp < normsq1 - 1e-14
        or normsqp < normsq2 - 1e-14
        or normsqm < normsq1 - 1e-14
        or normsqm < normsq2 - 1e-14
    ):
        raise ValueError("Lattice vectors are not of minimal length")
    # Rough estimate
    kpar -= b1 * np.round((kpar @ b1) / normsq1)
    kpar -= b2 * np.round((kpar @ b2) / normsq2)
    # todo: precise
    options = kpar + la.cube(2, 1) @ b
    for option in options:
        if option @ option < kpar @ kpar:
            kpar = option
    if n == 0 or np.array_equal(kpar, kparstart):
        return kpar
    return firstbrillouin2d(kpar, b, n - 1)


def firstbrillouin3d(kpar, b, n=2):
    """Map wave vector to first Brillouin zone in 3D.

    The reduction to the first Brillouin zone is first approximated roughly. From this
    approximated vector and its 26 neighbours, the shortest one is picked. As a
    sufficient approximation is not guaranteed (especially for extreme geometries),
    this process is iterated `n` times.

    Args:
        kpar (1d-array): parallel wave vector
        b (2d-array): reciprocal lattice vectors
        n (int): number of iterations

    Returns:
        (1d-array)
    """
    kparstart = kpar
    b1 = b[0, :]
    b2 = b[1, :]
    b3 = b[2, :]
    normsq1 = b1 @ b1
    normsq2 = b2 @ b2
    normsq3 = b3 @ b3
    # todo: Minimal length
    # Rough estimate
    kpar -= b1 * np.round((kpar @ b1) / normsq1)
    kpar -= b2 * np.round((kpar @ b2) / normsq2)
    kpar -= b3 * np.round((kpar @ b3) / normsq3)
    # todo: precise
    options = kpar + la.cube(3, 1) @ b
    for option in options:
        if option @ option < kpar @ kpar:
            kpar = option
    if n == 0 or np.array_equal(kpar, kparstart):
        return kpar
    return firstbrillouin3d(kpar, b, n - 1)



r"""Special (mathematical) functions.

.. currentmodule:: treams.special

Special mathematical functions used in :mod:`treams`. Some functions are reexported from
:py:mod:`scipy.special`. Most functions are available as Numpy universal functions
(:py:class:`numpy.ufunc`) or as generalized universal functions
(:ref:`c-api.generalized-ufuncs`).

Available functions
===================

Bessel and Hankel functions, with their spherical counterparts, derivatives
---------------------------------------------------------------------------

.. autosummary::
   :toctree:

   hankel1_d
   hankel2_d
   jv_d
   yv_d
   spherical_hankel1
   spherical_hankel2
   spherical_hankel1_d
   spherical_hankel2_d


Those functions are just reexported from Scipy. So, one only needs to import this
subpackage within treams.

+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.hankel1`\(v, z[, out])            | Hankel function of the  |
|                                                            | first kind.             |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.hankel2`\(v, z[, out])            | Hankel function of the  |
|                                                            | second kind.            |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.jv`\(v, z[, out])                 | Bessel function of the  |
|                                                            | first kind of real      |
|                                                            | order and complex       |
|                                                            | argument.               |
+------------------------------------------------------------+-------------------------+
| :py:data:`~scipy.special.yv`\(v, z[, out])                 | Bessel function of the  |
|                                                            | second kind of real     |
|                                                            | order and complex       |
|                                                            | argument.               |
+------------------------------------------------------------+-------------------------+
| | :py:func:`spherical_jn <scipy.special.spherical_jn>`\(n, | Spherical Bessel        |
|   z[, derivative])                                         | function of the first   |
|                                                            | kind or its derivative. |
+------------------------------------------------------------+-------------------------+
| | :py:func:`spherical_yn <scipy.special.spherical_yn>`\(n, | Spherical Bessel        |
|   z[, derivative])                                         | function of the second  |
|                                                            | kind or its derivative. |
+------------------------------------------------------------+-------------------------+

Those functions just wrap Scipy functions with special optional arguments to be able to
analogously access them like their non-spherical counterparts:

.. autosummary::
   :toctree:

   spherical_jn_d
   spherical_yn_d


Scipy functions with enhanced domain
------------------------------------

.. autosummary::
   :toctree:

   sph_harm
   lpmv


Integrals for the Ewald summation
---------------------------------

.. autosummary::
   :toctree:

   incgamma
   intkambe


Wigner d- and Wigner D-matrix elements
--------------------------------------

.. autosummary::
   :toctree:

   wignersmalld
   wignerd


Wigner 3j-symbols
-----------------

.. autosummary::
   :toctree:

   wigner3j


Vector wave functions
--------------------------

.. autosummary::
   :toctree:

   pi_fun
   tau_fun

Spherical waves and translation coefficients

.. autosummary::
   :toctree:

   vsh_X
   vsh_Y
   vsh_Z
   vsw_M
   vsw_N
   vsw_A
   vsw_rM
   vsw_rN
   vsw_rA
   tl_vsw_A
   tl_vsw_B
   tl_vsw_rA
   tl_vsw_rB


Cylindrical waves

.. autosummary::
   :toctree:

   vcw_M
   vcw_N
   vcw_A
   vcw_rM
   vcw_rN
   vcw_rA
   tl_vcw
   tl_vcw_r


Plane waves

.. autosummary::
   :toctree:

   vpw_M
   vpw_N
   vpw_A


Coordinate system transformations
---------------------------------

.. autosummary::
   :toctree:

   car2cyl
   car2sph
   cyl2car
   cyl2sph
   sph2car
   sph2cyl
   vcar2cyl
   vcar2sph
   vcyl2car
   vcyl2sph
   vsph2car
   vsph2cyl
   car2pol
   pol2car
   vcar2pol
   vpol2car


Cython module
-------------

.. autosummary::
   :toctree:
   :template: cython-module

   cython_special
"""

from scipy.special import (  # noqa: F401
    hankel1,
    hankel2,
    jv,
    spherical_jn,
    spherical_yn,
    yv,
)

from treams.special import _gufuncs, _ufuncs  # noqa: F401
from treams.special._gufuncs import *  # noqa: F401, F403
from treams.special._ufuncs import *  # noqa: F401, F403


def spherical_jn_d(n, z):
    """Derivative of the spherical Bessel function of the first kind.

    This is simply a wrapper for `scipy.special.spherical_jn(n, z, True)`, see
    :py:func:`scipy.special.spherical_jn`. It's here to have a consistent way of
    calling the derivative of a (spherical) Bessel or Hankel function.

    Args:
        n (int, array_like): Order
        z (float or complex, array_like): Argument

    Returns:
        float or complex

    References:
        - `DLMF 10.47 <https://dlmf.nist.gov/10.47>`_
        - `DLMF 10.51 <https://dlmf.nist.gov/10.51>`_
    """
    return spherical_jn(n, z, True)


def spherical_yn_d(n, z):
    """Derivative of the spherical Bessel function of the second kind.

    This is simply a wrapper for `scipy.special.spherical_yn(n, z, True)`, see
    :py:func:`scipy.special.spherical_jn`. It's here to have a consistent way of
    calling the derivative of a (spherical) Bessel or Hankel function.

    Args:
        n (int, array_like): Order
        z (float or complex, array_like): Argument

    Returns:
        float or complex

    References:
        - `DLMF 10.47 <https://dlmf.nist.gov/10.47>`_
        - `DLMF 10.51 <https://dlmf.nist.gov/10.51>`_
    """
    return spherical_yn(n, z, True)



import numpy as np

from treams import misc


class Material:
    r"""Material definition.

    The material properties are defined in the frequency domain through scalar values
    for permittivity :math:`\epsilon`, permeability :math:`\mu`, and the chirality
    parameter  :math:`\kappa`. Materials are, thus, assumed to be linear,
    time-invariant, homogeneous, isotropic, and local. Also, it is assumed that they
    have no gain. The relation of the electric and magnetic fields is defined by

    .. math::

        \begin{pmatrix}
            \frac{1}{\epsilon_0} \boldsymbol D \\
            c \boldsymbol B
        \end{pmatrix}
        =
        \begin{pmatrix}
            \epsilon & \mathrm i \kappa \\
            -\mathrm i \kappa & \mu
        \end{pmatrix}
        \begin{pmatrix}
            \boldsymbol E \\
            Z_0 \boldsymbol H
        \end{pmatrix}

    for a fixed vacuum wave number :math:`k_0` and spatial position
    :math:`\boldsymbol r` with :math:`\epsilon_0` the vacuum permittivity, :math:`c` the
    speed of light in vacuum, and :math:`Z_0` the vacuum impedance.

    Args:
        epsilon (optional, complex): Relative permittivity. Defaults to 1.
        mu (optional, complex): Relative permeability. Defaults to 1.
        kappa (optional, complex): Chirality parameter. Defaults to 0.
    """

    def __init__(self, epsilon=1, mu=1, kappa=0):
        """Initialization."""
        if isinstance(epsilon, Material):
            epsilon, mu, kappa = epsilon()
        elif isinstance(epsilon, (tuple, list, np.ndarray)):
            if len(epsilon) == 0:
                epsilon = 1
            elif len(epsilon) == 1:
                epsilon = epsilon[0]
            elif len(epsilon) == 2:
                epsilon, mu = epsilon
            elif len(epsilon) == 3:
                epsilon, mu, kappa = epsilon
            else:
                raise ValueError("invalid material definition")
        self._epsilon = epsilon
        self._mu = mu
        self._kappa = kappa

    @property
    def epsilon(self):
        """Relative permittivity.

        Returns:
            float or complex
        """
        return self._epsilon

    @property
    def mu(self):
        """Relative permeability.

        Returns:
            float or complex
        """
        return self._mu

    @property
    def kappa(self):
        """Chirality parameter.

        Returns:
            float or complex
        """
        return self._kappa

    def __iter__(self):
        """Iterator for a tuple containing the material parameters.

        Useful for unpacking the material parameters into a function that takes these
        parameters separately, e.g. ``foo(*material)``.
        """
        return iter((self.epsilon, self.mu, self.kappa))

    @classmethod
    def from_n(cls, n=1, impedance=None, kappa=0):
        r"""Create material from refractive index and relative impedance.

        This function calculates the relative permeability and permittivity with
        :math:`\epsilon = \frac{n}{Z}` and :math:`\mu = nZ`. The chirality parameter is
        considered separately.

        Note:
            The refractive index is defined independently from the chirality parameter
            here. For an alternative definition of the refractive index, see also
            :func:`Material.from_nmp`.

        Args:
            n (complex, optional): Refractive index. Defaults to 1.
            impedance(complex, optional): Relative impedance. Defaults to the inverse
                of the refractive index.
            kappa (complex, optional): Chirality parameter. Defaults to 0.

        Returns:
            Material
        """
        impedance = 1 / n if impedance is None else impedance
        epsilon = n / impedance
        mu = n * impedance
        return cls(epsilon, mu, kappa)

    @classmethod
    def from_nmp(cls, ns=(1, 1), impedance=None):
        r"""Create material from refractive indices of both helicities.

        This function calculates the relative permeability and permittivity and the
        chirality parameter with :math:`\epsilon = \frac{n_+ + n_-}{2Z}`,
        :math:`\mu = \frac{(n_+ + n_-)Z}{2}` and :math:`\mu = \frac{(n_+ - n_-)}{2}`.

        Note:
            Two refractive indices are defined here that depend on the chirality
            parameter. For an alternative definition of the refractive index, see also
            :func:`Material.from_n`.

        Args:
            ns ((2,)-array-like, optional): Negative and positive helicity refractive
                index. Defaults to (1, 1).
            impedance(complex, optional): Relative impedance. Defaults to the inverse of
                the refractive index.

        Returns:
            Material
        """
        n = sum(ns) * 0.5
        kappa = (ns[1] - ns[0]) * 0.5
        return cls.from_n(n, impedance, kappa)

    @property
    def n(self):
        r"""Refractive index.

        The refractive index is defined by :math:`n = \sqrt{\epsilon \mu}`, with an
        enforced non-negative imaginary part.

        Note:
            The refractive index returned is independent from the chirality parameter.
            For an alternative definition of the refractive index, see also
            :func:`Material.nmp`.

        Returns:
            complex
        """
        n = np.sqrt(self.epsilon * self.mu)
        if n.imag < 0:
            n = -n
        return n

    @property
    def nmp(self):
        r"""Refractive indices of both helicities.

        The refractive indices are defined by
        :math:`n_\pm = \sqrt{\epsilon \mu} \pm \kappa`, with an enforced non-negative
        imaginary part.

        Note:
            The refractive indices returned depend on the chirality parameter. For an
            alternative definition of the refractive index, see also :func:`Material.n`.

        Returns:
            tuple
        """
        return misc.refractive_index(self.epsilon, self.mu, self.kappa)

    @property
    def impedance(self):
        r"""Relative impedance.

        The relative impedance is defined by :math:`Z = \sqrt{\frac{\epsilon}{\mu}}`.

        Returns:
            complex
        """
        return np.sqrt(self.mu / self.epsilon)

    def __call__(self):
        """Return a tuple containing all material parameters.

        Returns:
            tuple
        """
        return self.epsilon, self.mu, self.kappa

    def __eq__(self, other):
        """Compare material parameters.

        Materials are considered equal, when all material parameters are equal. Also,
        compares with objects that contain at most three values.

        Returns:
            bool
        """
        if other is None:
            return False
        if not isinstance(other, Material):
            other = Material(*other)
        return (
            self.epsilon == other.epsilon
            and self.mu == other.mu
            and self.kappa == other.kappa
        )

    @property
    def ischiral(self):
        """Test if the material is chiral.

        Returns:
            bool
        """
        return self.kappa != 0

    @property
    def isreal(self):
        """Test if the material has purely real parameters.

        Returns:
            bool
        """
        return all(i.imag == 0 for i in self)

    def __str__(self):
        """All three material parameters.

        Returns:
            str
        """
        return "(" + ", ".join([str(i) for i in self()]) + ")"

    def __repr__(self):
        """Representation that allows recreating the object.

        Returns:
            str
        """
        return self.__class__.__name__ + str(self)

    def ks(self, k0):
        """Return the wave numbers in the medium for both polarizations.

        The first value corresponds to negative helicity and the second to positive
        helicity. For achiral materials where parity polarizations can be used both
        values are equal.

        Args:
            k0 (float): Wave number in vacuum.

        Returns:
            tuple
        """
        return k0 * self.nmp

    def krhos(self, k0, kz, pol):
        r"""The (cylindrically) radial part of the wave vector.

        The cylindrically radial part is defined by :math:`k_\rho = \sqrt(k^2 - k_z^2)`.
        In case of chiral materials :math:`k` and so :math`k_\rho` depends on the
        polarization. The returned values have non-negative imaginary parts.

        Args:
            k0 (float): Wave number in vacuum.
            kz (float, array-like): z-component of the wave vector
            pol (int, array-like): Polarization indices
                (:ref:`params:Polarizations`).

        Returns:
            complex, array-like
        """
        ks = self.ks(k0)[pol]
        return misc.wave_vec_z(kz, 0, ks)

    def kzs(self, k0, kx, ky, pol):
        r"""The z-component of the wave vector.

        The z-component of the wave vector is defined by
        :math:`k_z = \sqrt(k^2 - k_x^2 - k_y^2)`. In case of chiral materials :math:`k`
        and so :math`k_z` depends on the polarization. The returned values have
        non-negative imaginary parts.

        Args:
            k0 (float): Wave number in vacuum.
            kx (float, array-like): x-component of the wave vector
            ky (float, array-like): y-component of the wave vector
            pol (int, array-like): Polarization indices
                (:ref:`params:Polarizations`).

        Returns:
            complex, array-like
        """
        ks = self.ks(k0)[pol]
        return misc.wave_vec_z(kx, ky, ks)



import matplotlib.pyplot as plt
import numpy as np

import treams

k0 = 2 * np.pi / 700
materials = [treams.Material(-16.5 + 1j), treams.Material()]
lmax = mmax = 3
radii = [65, 55]
positions = [[-40, -50, 0], [40, 50, 0]]
lattice = treams.Lattice(300)
kz = 0.005

sphere = treams.TMatrix.sphere(lmax, k0, radii[0], materials)
chain = sphere.latticeinteraction.solve(lattice, kz)
bmax = 3.1 * lattice.reciprocal
cwb = treams.CylindricalWaveBasis.diffr_orders(kz, mmax, lattice, bmax)
chain_tmc = treams.TMatrixC.from_array(chain, cwb)

cylinder = treams.TMatrixC.cylinder(np.unique(cwb.kz), mmax, k0, radii[1], materials)

cluster = treams.TMatrixC.cluster([chain_tmc, cylinder], positions).interaction.solve()
inc = treams.plane_wave(
    [np.sqrt(k0 ** 2 - kz ** 2), 0, kz],
    [np.sqrt(0.5), np.sqrt(0.5)],
    k0=chain.k0,
    material=chain.material,
)
sca = cluster @ inc.expand(cluster.basis)

grid = (
    np.mgrid[-300:300:61j, 0:1, -150:150:31j].squeeze().transpose((1, 2, 0))[:, :-1, :]
)
ez = np.zeros_like(grid[..., 0], complex)
valid = cluster.valid_points(grid, radii)
ez[valid] = (inc.efield(grid[valid]) + sca.efield(grid[valid]))[..., 2]

ez = np.concatenate(
    (
        np.real(ez.T * np.exp(-1j * kz * lattice[...])),
        ez.T.real,
        np.real(ez.T * np.exp(1j * kz * lattice[...])),
    )
)
zaxis = np.concatenate((grid[0, :, 2] - 300, grid[0, :, 2], grid[0, :, 2] + 300,))

fig, ax = plt.subplots(figsize=(10, 20))
pcm = ax.pcolormesh(grid[:, 0, 0], zaxis, ez, shading="nearest", vmin=-1, vmax=1,)
cb = plt.colorbar(pcm)
cb.set_label("$E_z$")
ax.set_xlabel("x (nm)")
ax.set_ylabel("z (nm)")
ax.set_aspect("equal")
fig.show()



import matplotlib.pyplot as plt
import numpy as np

import treams

k0s = 2 * np.pi * np.linspace(1 / 10_000, 1 / 350, 200)
material_slab = 3
thickness = 10
material_sphere = (4 + 0.1j, 1, 0.05)
lattice = treams.Lattice.square(500)
radius = 100
lmax = 3
az = 210

res = []
for i, k0 in enumerate(k0s):
    kpar = [0, 0]
    spheres = treams.TMatrix.sphere(
        lmax, k0, radius, [material_sphere, 1]
    ).latticeinteraction.solve(lattice, kpar)

    pwb = treams.PlaneWaveBasisByComp.diffr_orders(kpar, lattice, 0.02)
    plw = treams.plane_wave(kpar, [1, 0, 0], k0=k0, basis=pwb, material=1)
    slab = treams.SMatrices.slab(thickness, pwb, k0, [1, material_slab, 1])
    dist = treams.SMatrices.propagation([0, 0, radius], pwb, k0, 1)
    array = treams.SMatrices.from_array(spheres, pwb)
    total = treams.SMatrices.stack([slab, dist, array, dist])
    x, _ = total.bands_kz(az)
    x = x * az / np.pi
    sel = x[np.abs(np.imag(x)) < 0.1]
    res.append(sel)


fig, ax = plt.subplots()
for k0, sel in zip(k0s, res):
    ax.scatter(sel.real, len(sel) * [299792.458 * k0 / (2 * np.pi)], 0.2, c="C0")
    ax.scatter(sel.imag, len(sel) * [299792.458 * k0 / (2 * np.pi)], 0.2, c="C1")
ax.set_xlabel("$k_z a_z / \\pi$")
ax.set_ylabel("Frequency (THz)")
ax.set_xlim([-1, 1])
ax.set_ylim(ymin=0)
ax.legend(["$Real$", "$Imag$"])
fig.show()



import matplotlib.pyplot as plt
import numpy as np

import treams

k0 = 2 * np.pi / 700
materials = [treams.Material(-16.5 + 1j), treams.Material()]
lmax = mmax = 3
radii = [75, 65]
positions = [[-30, 0, -75], [30, 0, 75]]
lattice = treams.Lattice(300)
kz = 0.005

spheres = [treams.TMatrix.sphere(lmax, k0, r, materials) for r in radii]
chain = treams.TMatrix.cluster(spheres, positions).latticeinteraction.solve(lattice, kz)

bmax = 3.1 * lattice.reciprocal
cwb = treams.CylindricalWaveBasis.diffr_orders(kz, mmax, lattice, bmax, 2, positions)
chain_tmc = treams.TMatrixC.from_array(chain, cwb)

inc = treams.plane_wave(
    [np.sqrt(k0 ** 2 - kz ** 2), 0, kz],
    [np.sqrt(0.5), np.sqrt(0.5)],
    k0=chain.k0,
    material=chain.material,
)
sca = chain_tmc @ inc.expand(chain_tmc.basis)

grid = (
    np.mgrid[-300:300:61j, 0:1, -150:150:31j].squeeze().transpose((1, 2, 0))[:, :-1, :]
)
ez = np.zeros_like(grid[..., 0], complex)
valid = chain_tmc.valid_points(grid, radii)
ez[valid] = (inc.efield(grid[valid]) + sca.efield(grid[valid]))[..., 2]
ez[~valid] = np.nan
sca = chain @ inc.expand(chain.basis)
valid = ~valid & chain.valid_points(grid, radii)
vals = []
for i, r in enumerate(grid[valid]):
    swb = treams.SphericalWaveBasis.default(1, positions=[r])
    field = sca.expandlattice(basis=swb).efield(r)
    vals.append(inc.efield(r)[2] + field[2])
ez[valid] = vals

ez = np.concatenate(
    (
        np.real(ez.T * np.exp(-1j * kz * lattice[...])),
        ez.T.real,
        np.real(ez.T * np.exp(1j * kz * lattice[...])),
    )
)
zaxis = np.concatenate((grid[0, :, 2] - 300, grid[0, :, 2], grid[0, :, 2] + 300,))

fig, ax = plt.subplots(figsize=(10, 20))
pcm = ax.pcolormesh(grid[:, 0, 0], zaxis, ez, shading="nearest", vmin=-1, vmax=1,)
cb = plt.colorbar(pcm)
cb.set_label("$E_z$")
ax.set_xlabel("x (nm)")
ax.set_ylabel("z (nm)")
ax.set_aspect("equal")
fig.show()



"""Configuration for pytest.

It's needed here to add the option `--runslow`, which is mainly used in the lattice
subpackage.
"""
import pytest


def pytest_addoption(parser):
    """Add option '--runslow' and '--rungmsh'."""
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )
    parser.addoption(
        "--rungmsh", action="store_true", default=False, help="run tests needing gmsh"
    )


def pytest_configure(config):
    """Add marker 'slow' and 'gmsh'."""
    config.addinivalue_line("markers", "gmsh: test needs gmsh")
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    """Skip tests without necessary options."""
    if not config.getoption("--runslow"):
        # --runslow not given in cli: skip slow tests
        skip_slow = pytest.mark.skip(reason="need --runslow option to run")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)
    if not config.getoption("--rungmsh"):
        # --rungmsh not given in cli: skip gmsh tests
        skip_slow = pytest.mark.skip(reason="need --rungmsh option to run")
        for item in items:
            if "gmsh" in item.keywords:
                item.add_marker(skip_slow)



import numpy as np
import pytest

import treams


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


class TestSWB:
    def test_init_empty(self):
        b = treams.SphericalWaveBasis([])
        assert b.l.size == 0 and b.m.size == 0 and b.pol.size == 0 and b.pidx.size == 0

    def test_init_numpy(self):
        b = treams.SphericalWaveBasis(np.array([[1, 0, 0]]), [0, 1, 0])
        assert (
            np.all(b.l == [1])
            and np.all(b.m == [0])
            and np.all(b.pol == [0])
            and np.all(b.pidx == [0])
        )

    def test_init_duplicate(self):
        b = treams.SphericalWaveBasis([[0, 1, 0, 0], [0, 1, 0, 0]])
        assert (
            np.all(b.l == [1])
            and np.all(b.m == [0])
            and np.all(b.pol == [0])
            and np.all(b.pidx == [0])
        )

    def test_init_invalid_shape(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[0, 0], [0, 0]])

    def test_init_invalid_positions(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[1, 0, 0]], [1, 2])

    def test_init_non_int_value(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[1.1, 0, 0]])

    def test_init_non_natural_l(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[0, 0, 0]])

    def test_init_non_too_large_m(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[1, -2, 0]])

    def test_init_invalid_pol(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[1, 0, 2]])

    def test_init_unsecified_positions(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis([[1, 1, 0, 0]])

    def test_property_positions(self):
        a = np.array([[1, 2, 3]])
        b = treams.SphericalWaveBasis([[1, 0, 0]], a)
        assert (a == b.positions).all()

    def test_repr(self):
        b = treams.SphericalWaveBasis(
            [[0, 1, 0, 0], [1, 1, 0, 0]], [[0.0, 0, 0], [1, 0, 0]]
        )
        assert (
            repr(b)
            == """SphericalWaveBasis(
    pidx=[0 1],
    l=[1 1],
    m=[0 0],
    pol=[0 0],
    positions=[[0. 0. 0.], [1. 0. 0.]],
)"""
        )

    def test_getitem_plms(self):
        b = treams.SphericalWaveBasis([[0, 2, -1, 1]])
        assert b.plms == ([0], [2], [-1], [1])

    def test_getitem_lms(self):
        b = treams.SphericalWaveBasis([[2, -1, 1]])
        assert b.lms == ([2], [-1], [1])

    def test_getitem_invalid_index(self):
        b = treams.SphericalWaveBasis([[1, 0, 0]])
        with pytest.raises(AttributeError):
            b.fail

    def test_getitem_int(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert b[1] == (0, 1, 0, 1)

    def test_getitem_tuple(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert (np.array(b[()]) == ([0, 0], [1, 1], [0, 0], [0, 1])).all()

    def test_getitem_slice(self):
        a = treams.SphericalWaveBasis([[1, 0, 0]])
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert a == b[:1]

    def test_default(self):
        a = treams.SphericalWaveBasis.default(2, 2, [[0, 0, 0], [1, 0, 0]])
        b = treams.SphericalWaveBasis(
            zip(
                16 * [0] + 16 * [1],
                2 * (6 * [1] + 10 * [2]),
                2 * [-1, -1, 0, 0, 1, 1, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2],
                16 * [1, 0],
            ),
            [[0, 0, 0], [1, 0, 0]],
        )
        assert a == b

    def test_ebcm(self):
        a = treams.SphericalWaveBasis.ebcm(2, 2, positions=[[0, 0, 0], [1, 0, 0]])
        b = treams.SphericalWaveBasis(
            zip(
                16 * [0] + 16 * [1],
                2 * ([2, 2] + 3 * [1, 1, 2, 2] + [2, 2]),
                2 * [-2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2],
                16 * [1, 0],
            ),
            [[0, 0, 0], [1, 0, 0]],
        )
        assert a == b

    def test_ebcm_mmax(self):
        a = treams.SphericalWaveBasis.ebcm(2, mmax=1)
        b = treams.SphericalWaveBasis(
            zip(
                3 * [1, 1, 2, 2], [-1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1], 6 * [1, 0],
            ),
        )
        assert a == b

    def test_defaultlmax(self):
        assert treams.SphericalWaveBasis.defaultlmax(60, 2) == 3

    def test_defaultlmax_fail(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis.defaultlmax(1)

    def test_defaultdim(self):
        assert treams.SphericalWaveBasis.defaultdim(3, 2) == 60

    def test_defaultdim_fail(self):
        with pytest.raises(ValueError):
            treams.SphericalWaveBasis.defaultdim(1, -1)

    def test_property_isglobal_true(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert b.isglobal

    def test_property_isglobal_false(self):
        b = treams.SphericalWaveBasis(
            [[0, 1, 0, 0], [1, 1, 0, 0]], [[0, 0, 0], [1, 0, 0]]
        )
        assert not b.isglobal

    def test_from_iterable(self):
        a = treams.SphericalWaveBasis.default(1)
        b = treams.SphericalWaveBasis.default(2)
        assert a & b == a

    def test_neq(self):
        b = treams.SphericalWaveBasis.default(1)
        assert not b == []


class TestCWB:
    def test_init_empty(self):
        b = treams.CylindricalWaveBasis([])
        assert b.kz.size == 0 and b.m.size == 0 and b.pol.size == 0 and b.pidx.size == 0

    def test_init_numpy(self):
        b = treams.CylindricalWaveBasis(np.array([[0.5, 0, 0]]), [0, 1, 0])
        assert (
            np.all(b.kz == [0.5])
            and np.all(b.m == [0])
            and np.all(b.pol == [0])
            and np.all(b.pidx == [0])
        )

    def test_init_duplicate(self):
        b = treams.CylindricalWaveBasis([[0, 1, 0, 0], [0, 1, 0, 0]])
        assert (
            np.all(b.kz == [1])
            and np.all(b.m == [0])
            and np.all(b.pol == [0])
            and np.all(b.pidx == [0])
        )

    def test_init_invalid_shape(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis([[0, 0], [0, 0]])

    def test_init_invalid_positions(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis([[1, 0, 0]], [1, 2])

    def test_init_non_int_value(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis([[1, 0.2, 0]])

    def test_init_invalid_pol(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis([[1, 0, 2]])

    def test_init_unsecified_positions(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis([[1, 1, 0, 0]])

    def test_property_positions(self):
        a = np.array([[1, 2, 3]])
        b = treams.CylindricalWaveBasis([[1, 0, 0]], a)
        assert (a == b.positions).all()

    def test_repr(self):
        b = treams.CylindricalWaveBasis(
            [[0, 1, 0, 0], [1, 1, 0, 0]], [[0.0, 0, 0], [1, 0, 0]]
        )
        assert (
            repr(b)
            == """CylindricalWaveBasis(
    pidx=[0 1],
    kz=[1. 1.],
    m=[0 0],
    pol=[0 0],
    positions=[[0. 0. 0.], [1. 0. 0.]],
)"""
        )

    def test_getitem_pzms(self):
        b = treams.CylindricalWaveBasis([[0, 2, -1, 1]])
        assert b.pzms == ([0], [2], [-1], [1])

    def test_getitem_zms(self):
        b = treams.CylindricalWaveBasis([[2, -1, 1]])
        assert b.zms == ([2], [-1], [1])

    def test_getitem_invalid_index(self):
        b = treams.CylindricalWaveBasis([[1, 0, 0]])
        with pytest.raises(AttributeError):
            b.fail

    def test_getitem_int(self):
        b = treams.CylindricalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert b[1] == (0, 1, 0, 1)

    def test_getitem_tuple(self):
        b = treams.CylindricalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert (np.array(b[()]) == ([0, 0], [1, 1], [0, 0], [0, 1])).all()

    def test_getitem_slice(self):
        a = treams.CylindricalWaveBasis([[1, 0, 0]])
        b = treams.CylindricalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert a == b[:1]

    def test_default(self):
        a = treams.CylindricalWaveBasis.default(
            [0.3, -0.2], 2, 2, [[0, 0, 0], [1, 0, 0]]
        )
        b = treams.CylindricalWaveBasis(
            zip(
                20 * [0] + 20 * [1],
                2 * (10 * [0.3] + 10 * [-0.2]),
                4 * [-2, -2, -1, -1, 0, 0, 1, 1, 2, 2],
                20 * [1, 0],
            ),
            [[0, 0, 0], [1, 0, 0]],
        )
        assert a == b

    def test_defaultmmax(self):
        assert treams.CylindricalWaveBasis.defaultmmax(112, 4, 2) == 3

    def test_defaultmmax_fail(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis.defaultmmax(1)

    def test_defaultdim(self):
        assert treams.CylindricalWaveBasis.defaultdim(3, 2, 4) == 120

    def test_defaultdim_fail(self):
        with pytest.raises(ValueError):
            treams.CylindricalWaveBasis.defaultdim(1, -1)

    def test_property_isglobal_true(self):
        b = treams.CylindricalWaveBasis([[1, 0, 0], [1, 0, 1]])
        assert b.isglobal

    def test_property_isglobal_false(self):
        b = treams.CylindricalWaveBasis(
            [[0, 1, 0, 0], [1, 1, 0, 0]], [[0, 0, 0], [1, 0, 0]]
        )
        assert not b.isglobal

    def test_from_iterable(self):
        a = treams.CylindricalWaveBasis.default(0, 1)
        b = treams.CylindricalWaveBasis.default(0, 2)
        assert a & b == a

    def test_neq(self):
        b = treams.CylindricalWaveBasis.default(0, 1)
        assert not b == []

    def test_diffr_orders(self):
        a = treams.CylindricalWaveBasis.diffr_orders(0.1, 1, 2 * np.pi, 1.5)
        b = treams.CylindricalWaveBasis(
            zip(
                *[
                    6 * [-0.9] + 6 * [0.1] + 6 * [1.1],
                    3 * [-1, -1, 0, 0, 1, 1],
                    18 * [1, 0],
                ]
            )
        )
        assert (
            a == b
            and a.lattice == treams.Lattice(2 * np.pi)
            and a.kpar == [np.nan, np.nan, 0.1]
        )


class TestPWBUV:
    def test_init_empty(self):
        b = treams.PlaneWaveBasisByUnitVector([])
        assert b.qx.size == 0 and b.qy.size == 0 and b.qz.size == 0 and b.pol.size == 0

    def test_init_numpy(self):
        b = treams.PlaneWaveBasisByUnitVector(np.array([[0.1, 0.2, 0.2, 0]]))
        assert (
            np.all(b.qx == [1 / 3])
            and np.all(b.qy == [2 / 3])
            and np.all(b.qz == [2 / 3])
            and np.all(b.pol == [0])
        )

    def test_init_duplicate(self):
        b = treams.PlaneWaveBasisByUnitVector([[0.1, 0.2, 0.2, 0], [0.1, 0.2, 0.2, 0]])
        assert (
            np.all(b.qx == [1 / 3])
            and np.all(b.qy == [2 / 3])
            and np.all(b.qz == [2 / 3])
            and np.all(b.pol == [0])
        )

    def test_init_invalid_shape(self):
        with pytest.raises(ValueError):
            treams.PlaneWaveBasisByUnitVector([[0, 0], [0, 0]])

    def test_init_invalid_pol(self):
        with pytest.raises(ValueError):
            treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 2]])

    def test_repr(self):
        b = treams.PlaneWaveBasisByUnitVector([[0.0, 1.0, 0.0, 0], [1, 0, 0, 0]])
        assert (
            repr(b)
            == """PlaneWaveBasisByUnitVector(
    qx=[0. 1.],
    qy=[1. 0.],
    qz=[0. 0.],
    pol=[0 0],
)"""
        )

    def test_getitem_xyzs(self):
        b = treams.PlaneWaveBasisByUnitVector([[0, 4, -3, 1]])
        assert b.xyzs == ([0], [0.8], [-0.6], [1])

    def test_getitem_xys(self):
        b = treams.PlaneWaveBasisByUnitVector([[0, 4, -3, 1]])
        assert b.xys == ([0], [0.8], [1])

    def test_getitem_invalid_index(self):
        b = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0]])
        with pytest.raises(AttributeError):
            b.fail

    def test_getitem_int(self):
        b = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0], [1, 0, 0, 1]])
        assert b[1] == (1, 0, 0, 1)

    def test_getitem_tuple(self):
        b = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0], [1, 0, 0, 1]])
        assert (np.array(b[()]) == ([1, 1], [0, 0], [0, 0], [0, 1])).all()

    def test_getitem_slice(self):
        a = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0]])
        b = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0], [1, 0, 0, 1]])
        assert a == b[:1]

    def test_default(self):
        a = treams.PlaneWaveBasisByUnitVector.default([0.3, -0.2, 0.1])
        b = treams.PlaneWaveBasisByUnitVector(
            [[0.3, -0.2, 0.1, 1], [0.3, -0.2, 0.1, 0]]
        )
        assert a == b

    def test_property_isglobal_true(self):
        b = treams.PlaneWaveBasisByUnitVector([])
        assert b.isglobal

    def test_from_iterable(self):
        a = treams.PlaneWaveBasisByUnitVector.default([0, 0, 1])
        b = treams.PlaneWaveBasisByUnitVector.default([[0, 0, 1], [0, 1, 0]])
        assert a & b == a

    def test_bycomp(self):
        a = treams.PlaneWaveBasisByComp.default([0, 1], "yz")
        b = treams.PlaneWaveBasisByUnitVector.default([0, 0, 1])
        assert b.bycomp(1, "yz") == a


class TestPWBC:
    def test_init_empty(self):
        b = treams.PlaneWaveBasisByComp([])
        assert b.kx.size == 0 and b.ky.size == 0 and b.kz is None and b.pol.size == 0

    def test_init_numpy(self):
        b = treams.PlaneWaveBasisByComp(np.array([[0.4, 0.2, 0]]), "zx")
        assert (
            np.all(b.kz == [0.4])
            and np.all(b.kx == [0.2])
            and np.all(b.ky is None)
            and np.all(b.pol == [0])
        )

    def test_init_duplicate(self):
        b = treams.PlaneWaveBasisByComp([[0.4, 0.2, 0], [0.4, 0.2, 0]])
        assert (
            np.all(b.kx == [0.4])
            and np.all(b.ky == [0.2])
            and np.all(b.kz is None)
            and np.all(b.pol == [0])
        )

    def test_init_invalid_shape(self):
        with pytest.raises(ValueError):
            treams.PlaneWaveBasisByComp([[0, 0], [0, 0]])

    def test_init_invalid_pol(self):
        with pytest.raises(ValueError):
            treams.PlaneWaveBasisByComp([[1, 0, 2]])

    def test_repr(self):
        b = treams.PlaneWaveBasisByComp([[0.0, 1.0, 0], [1, 1, 0]], "yz")
        assert (
            repr(b)
            == """PlaneWaveBasisByComp(
    ky=[0. 1.],
    kz=[1. 1.],
    pol=[0 0],
)"""
        )

    def test_from_iterable(self):
        a = treams.PlaneWaveBasisByComp.default([0, 0])
        b = treams.PlaneWaveBasisByComp.default([[0, 0], [0, 1]])
        assert a & b == a

    def test_getitem_int(self):
        b = treams.PlaneWaveBasisByComp([[1, 0, 0], [1, 0, 1]])
        assert b[1] == (1, 0, 1)

    def test_getitem_tuple(self):
        b = treams.PlaneWaveBasisByComp([[1, 0, 0], [1, 0, 1]])
        assert (np.array(b[()]) == ([1, 1], [0, 0], [0, 1])).all()

    def test_getitem_slice(self):
        a = treams.PlaneWaveBasisByComp([[1, 0, 0]])
        b = treams.PlaneWaveBasisByComp([[1, 0, 0], [1, 0, 0, 1]])
        assert a == b[:1]

    def test_byunitvector(self):
        a = treams.PlaneWaveBasisByUnitVector.default([0, 0, 1])
        b = treams.PlaneWaveBasisByComp.default([0, 1], "yz")
        assert b.byunitvector(1) == a

    def test_diffr_orders(self):
        lattice = treams.Lattice(2 * np.pi * np.eye(2))
        b = treams.PlaneWaveBasisByComp.diffr_orders([0, 0], lattice, 1)
        a = treams.PlaneWaveBasisByComp.default(
            [[0, 0], [0, 1], [1, 0], [-1, 0], [0, -1]]
        )
        assert a <= b and b <= a and b.lattice == lattice and b.kpar == [0, 0, np.nan]


class TestPhysicsArray:
    def test_init(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray(np.eye(2), basis=b)
        assert (p == np.eye(2)).all() and p.basis == b

    def test_type_error(self):
        with pytest.raises(TypeError):
            treams.PhysicsArray(np.eye(2), basis="fail")

    def test_lattice(self):
        b = treams.PlaneWaveBasisByComp.diffr_orders([0, 0], np.eye(2), 4)
        p = treams.PhysicsArray([1, 2], lattice=treams.Lattice(1, "x"), basis=b)
        assert p.lattice == treams.Lattice(1, "x")

    def test_matmul(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([[1, 1], [2, 2]], basis=b)
        x = p @ [1, 2]
        assert (x == [3, 6]).all() and x.basis == b

    def test_rmatmul(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([[1, 1], [2, 2]], basis=b)
        x = [1, 2] @ p
        assert (x == [5, 5]).all() and x.basis == b

    def test_changepoltype(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([[1, 0], [0, 1]], basis=b, poltype="helicity")
        x = p.changepoltype.apply_left()
        assert (
            x == np.sqrt(0.5) * np.array([[-1, 1], [1, 1]])
        ).all() and x.poltype == ("parity", "helicity",)

    def test_changepoltype_inv(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([[1, 0], [0, 1]], basis=b, poltype="helicity")
        x = p.changepoltype.apply_right()
        assert (
            x == np.sqrt(0.5) * np.array([[-1, 1], [1, 1]])
        ).all() and x.poltype == ("helicity", "parity",)

    def test_efield(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b, k0=1, poltype="helicity")
        x = p.efield([1, 0, 0])
        assert (
            np.abs(
                x
                - (
                    treams.special.vsph2car(
                        treams.special.vsw_rA(1, 0, 1, np.pi / 2, 0, [0, 1]),
                        [1, np.pi / 2, 0],
                    ).sum(0)
                )
                < 1e-15
            )
        ).all()

    def test_efield_inv(self):
        with pytest.raises(NotImplementedError):
            treams.PhysicsArray([0]).efield.apply_right([0, 0, 0])

    def test_expand(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b, k0=1, poltype="helicity")
        x = p.expand(b)
        assert (np.abs(x - [1, 0]) < 1e-14).all()

    def test_expand_inv(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b, k0=1, poltype="helicity")
        x = p.expand.eval_inv(b)
        assert (np.abs(x - np.eye(2)) < 1e-14).all()

    def test_expandlattice(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b, k0=1, poltype="helicity")
        x = p.expandlattice.eval(lattice=[[1, 0], [0, 1]], kpar=[0, 0])
        assert x.modetype == ("regular", "singular")

    def test_expandlattice_inv(self):
        with pytest.raises(NotImplementedError):
            treams.PhysicsArray([0]).expandlattice.eval_inv(lattice=1)

    def test_permute(self):
        b = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0], [1, 0, 0, 1]])
        c = treams.PlaneWaveBasisByUnitVector([[0, 1, 0, 0], [0, 1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b)
        assert p.permute.eval().basis == (c, b)

    def test_permute_inv(self):
        b = treams.PlaneWaveBasisByUnitVector([[1, 0, 0, 0], [1, 0, 0, 1]])
        c = treams.PlaneWaveBasisByUnitVector([[0, 1, 0, 0], [0, 1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b)
        assert p.permute.eval_inv().basis == (b, c)

    def test_rotate(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b)
        x = p.rotate.eval(1, 2, 3)
        y = np.diag([treams.special.wignerd(1, 0, 0, 1, 2, 3)] * 2)
        assert (np.abs(x - y) < 1e-14).all()

    def test_rotate_inv(self):
        b = treams.SphericalWaveBasis([[1, -1, 0], [1, 0, 0], [1, 1, 0]])
        p = treams.PhysicsArray([1, 0, 0], basis=b)
        x = p.rotate.eval_inv(1, 2, 3)
        y = treams.special.wignerd(1, [[-1], [0], [1]], [-1, 0, 1], 1, 2, 3)
        assert (np.abs(x - y.conj().T) < 1e-14).all()

    def test_translate(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b, k0=1)
        x = p.translate.eval([0, 0, 0])
        assert (np.abs(x - np.eye(2)) < 1e-14).all()

    def test_translate_inv(self):
        b = treams.SphericalWaveBasis([[1, 0, 0], [1, 0, 1]])
        p = treams.PhysicsArray([1, 0], basis=b, k0=1)
        x = p.translate.eval_inv([0, 0, 0])
        assert (np.abs(x - np.eye(2)) < 1e-14).all()



import copy

import numpy as np
import pytest

from treams import util


class TestAnnotationDict:
    def test_init_dict(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        assert a == {"a": 1, "b": 2}

    def test_init_tuple(self):
        a = util.AnnotationDict((("a", 1), ("b", 2)))
        assert a == {"a": 1, "b": 2}

    def test_init_anndict(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        b = util.AnnotationDict(a)
        assert a == b

    def test_kwargs(self):
        with pytest.warns(util.AnnotationWarning):
            a = util.AnnotationDict({"a": 0, "c": 3}, a=1, b=2)
        assert a == {"a": 1, "b": 2, "c": 3}

    def test_getitem(self):
        a = util.AnnotationDict({"a": 1})
        assert a["a"] == 1

    def test_delitem(self):
        a = util.AnnotationDict({"a": 1})
        del a["a"]
        assert a == {}

    def test_iter(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        assert ["a", "b"] == list(a)

    def test_len(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        assert len(a) == 2

    def test_repr(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        assert repr(a) == "AnnotationDict({'a': 1, 'b': 2})"

    def test_match(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        b = {"a": 1, "c": 3}
        assert a.match(b) is None

    def test_match_fail(self):
        a = util.AnnotationDict({"a": 1, "b": 2})
        b = {"a": 2, "c": 3}
        with pytest.warns(util.AnnotationWarning):
            a.match(b)


class TestAnnotationSequence:
    def test_init(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})
        assert a == ({"a": 1}, {"a": 2})

    def test_repr(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2}, mapping=dict)
        assert repr(a) == "AnnotationSequence({'a': 1}, {'a': 2})"

    def test_len(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})
        assert len(a) == 2

    def test_getitem_int(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})
        assert a[1] == {"a": 2}

    def test_getitem_tuple(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})[()]
        assert a == ({"a": 1}, {"a": 2}) and isinstance(a, tuple)

    def test_getitem_slice(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})
        assert a[1:] == ({"a": 2},) and isinstance(a, util.AnnotationSequence)

    def test_getitem_list(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})
        assert a[[1, 0, 1]] == ({"a": 2}, {"a": 1}, {"a": 2})

    def test_getitem_error(self):
        a = util.AnnotationSequence({"a": 1}, {"a": 2})
        with pytest.raises(TypeError):
            a["fail"]

    def test_update_warn_len(self):
        a = util.AnnotationSequence({})
        with pytest.warns(util.AnnotationWarning):
            a.update(({}, {}))

    def test_update_warn_overwrite(self):
        a = util.AnnotationSequence({"a": 1})
        with pytest.warns(util.AnnotationWarning):
            a.update(({"a": 2},))

    def test_update(self):
        a = util.AnnotationSequence({"a": 1}, {"b": 2})
        a.update(({"c": 3}, {"a": 4}))
        assert a == util.AnnotationSequence({"a": 1, "c": 3}, {"b": 2, "a": 4})

    def test_match_warn_overwrite(self):
        a = util.AnnotationSequence({"a": 1})
        with pytest.warns(util.AnnotationWarning):
            a.match(({"a": 2},))

    def test_match(self):
        a = util.AnnotationSequence({"a": 1})
        assert a.match(({"b": 2},)) is None

    def test_eq_error_len_get(self):
        assert not (util.AnnotationSequence({}) == 1)

    def test_eq_error_len(self):
        assert not (util.AnnotationSequence({}, {}) == (None,))

    def test_eq_false(self):
        assert not (
            util.AnnotationSequence({"a": 1}) == util.AnnotationSequence({"a": 2})
        )

    def test_eq_true(self):
        assert util.AnnotationSequence({"a": 1}) == util.AnnotationSequence({"a": 1})

    def test_add(self):
        assert util.AnnotationSequence({"a": 1}) + (
            {"b": 2},
        ) == util.AnnotationSequence({"a": 1}, {"b": 2})

    def test_radd(self):
        assert ({"a": 1},) + util.AnnotationSequence(
            {"b": 2}
        ) == util.AnnotationSequence({"a": 1}, {"b": 2})


class TestSequenceAsDict:
    def test_set(self):
        ann = util.AnnotationSequence({}, {})
        ann.as_dict = {"a": (1, 2)}
        assert ann.as_dict["a"] == (1, 2)

    def test_getitem(self):
        ann = util.AnnotationSequence({"a": 1}, {"a": 1})
        assert ann.as_dict["a"] == (1, 1)

    def test_getitem_missing(self):
        ann = util.AnnotationSequence({})
        with pytest.raises(KeyError):
            ann.as_dict["fail"]

    def test_setitem_delnone(self):
        ann = util.AnnotationSequence({"a": 1}, {})
        ann.as_dict["a"] = (1,)
        assert ann.as_dict["a"] == (None, 1)

    def test_delitem(self):
        ann = util.AnnotationSequence({"a": 1})
        del ann.as_dict["a"]
        assert ann == util.AnnotationSequence({})

    def test_delitem_error(self):
        ann = util.AnnotationSequence({})
        with pytest.raises(KeyError):
            del ann.as_dict["fail"]

    def test_iter(self):
        ann = util.AnnotationSequence({"a": 1}, {"a": 1, "b": 2})
        assert set(iter(ann.as_dict)) == {"a", "b"}

    def test_len(self):
        ann = util.AnnotationSequence({"a": 1, "b": 2})
        assert len(ann.as_dict) == 2

    def test_repr(self):
        ann = util.AnnotationSequence({"a": 1}, {"a": 1, "b": 2})
        assert repr(ann.as_dict) in (
            "SequenceAsDict({'a': (1, 1), 'b': (None, 2)})",
            "SequenceAsDict({'b': (None, 2), 'a': (1, 1)})",
        )


class TestAnnotatedArray:
    def test_init(self):
        arr = util.AnnotatedArray([1, 2, 3], a=(1,))
        assert (arr == [1, 2, 3]).all() and arr.ann == ({"a": 1},)

    def test_getattr_same(self):
        arr = util.AnnotatedArray([[0, 1], [2, 3], [4, 5]], a=(0, 0))
        assert arr.a == 0

    def test_getattr(self):
        arr = util.AnnotatedArray([[0, 1], [2, 3], [4, 5]], a=(0, 1))
        assert arr.a == (0, 1)

    def test_setattr(self):
        arr = util.AnnotatedArray([[0, 1], [2, 3], [4, 5]], a=(1,))
        arr.a = (0, 1)
        assert arr.a == (0, 1)

    def test_delattr(self):
        arr = util.AnnotatedArray([[0, 1], [2, 3], [4, 5]], a=(1,))
        del arr.a
        assert arr.ann == util.AnnotationSequence({}, {})

    def test_copy(self):
        a = util.AnnotatedArray([[0, 1], [2, 3], [4, 5]], a=(1,))
        b = copy.copy(a)
        assert (a == b).all() and a.ann == b.ann and (a is not b)

    def test_ufunc_out(self):
        a = util.AnnotatedArray([[2, 3], [4, 5]], a=(2, 1))
        b = util.AnnotatedArray([2, 4], a=(1,))
        x = util.AnnotatedArray([[-1, -1], [-1, -1]])
        y = util.AnnotatedArray([[-1, -1], [-1, -1]])
        np.divmod(a, b, out=(x, y))
        assert (
            (x == [[1, 0], [2, 1]]).all()
            and (y == [[0, 3], [0, 1]]).all()
            and x.ann == y.ann == a.ann
        )

    def test_bool(self):
        a = util.AnnotatedArray([0, 0], a=(0,))
        with pytest.raises(ValueError):
            bool(a)

    def test_bool(self):
        a = util.AnnotatedArray([1], a=(0,))
        assert bool(a)

    def test_ufunc_reduce(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        y = np.add.reduce(x)
        assert (y == [4, 6]).all() and y.ann == ({"a": 1},)

    def test_ufunc_at(self):
        arr = util.AnnotatedArray([[2, 3], [4, 5]], a=(2, 1))
        np.add.at(arr, 1, [1, 2])
        assert (arr == [[2, 3], [5, 7]]).all() and arr.a == (2, 1)

    def test_ufunc_outer(self):
        x = util.AnnotatedArray([[2, 3], [4, 5]], a=(2, 1))
        y = util.AnnotatedArray([10, -1], a=(3,), b=(4,))
        z = np.add.outer(x, y)
        assert (
            (z == [[[12, 1], [13, 2]], [[14, 3], [15, 4]]]).all()
            and z.a == (2, 1, 3)
            and z.b == (None, None, 4)
        )

    def test_gufunc(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        y = util.AnnotatedArray([[[1], [-1]], [[10], [-1]]], a=(3, 1, None), b=(4, 4))
        z = x @ y
        assert (z == [[[-1], [-1]], [[8], [26]]]).all() and z.ann == (
            {"a": 3},
            {"a": 2},
            {"b": 4},
        )

    def test_str(self):
        assert str(util.AnnotatedArray([0, 1], a=(1,))) == "[0 1]"

    def test_repr(self):
        assert (
            repr(util.AnnotatedArray([0, 1], a=(1,)))
            == """AnnotatedArray(
    [0, 1],
    AnnotationSequence(AnnotationDict({'a': 1})),
)"""
        )

    def test_getitem_slice(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        y = x[:, 0]
        assert (y == [1, 3]).all() and y.ann == ({"a": 2},)

    def test_getitem_none(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        y = x[:, None, 0]
        assert (y == [[1], [3]]).all() and y.ann == ({"a": 2}, {})

    def test_getitem_ellipsis(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        y = x[1, ...]
        assert (y == [3, 4]).all() and y.ann == ({"a": 1},)

    def test_getitem_fancy(self):
        x = util.AnnotatedArray([[[0, 1], [2, 3]], [[4, 5], [6, 7]]], a=(2, 1, 2))
        y = x[[True, False], :, [1, 0]]
        assert (y == [[1, 3], [0, 2]]).all() and y.ann == ({"a": 2}, {"a": 1})

    def test_getitem_ellipsis_implicit(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        y = x[1]
        assert (y == [3, 4]).all() and y.ann == ({"a": 1},)

    def test_setitem_slice(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(0, 1))
        x[:, 1] = util.AnnotatedArray([5, 6], a=(0,))
        assert (x == [[1, 5], [3, 6]]).all() and x.ann == ({"a": 0}, {"a": 1})

    def test_setitem_none(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        x[:, None, 1] = util.AnnotatedArray([[5], [6]], a=(2, -1))
        assert (x == [[1, 5], [3, 6]]).all() and x.ann == ({"a": 2}, {"a": 1})

    def test_setitem_ellipsis(self):
        x = util.AnnotatedArray([[1, 2], [3, 4]], a=(2, 1))
        x[..., 1] = util.AnnotatedArray([5, 6], a=(2,))
        assert (x == [[1, 5], [3, 6]]).all() and x.ann == ({"a": 2}, {"a": 1})

    def test_setitem_fancy(self):
        x = util.AnnotatedArray([[[0, 1], [2, 3]], [[4, 5], [6, 7]]], a=(1, 2, 1))
        x[[True, False], :, [1, 0]] = util.AnnotatedArray(
            [[-1, -2], [-3, -4]], a=(1, 2), b=(-1, -2)
        )
        assert (x == [[[-3, -1], [-4, -2]], [[4, 5], [6, 7]]]).all() and x.ann == (
            {"a": 1},
            {"a": 2},
            {"a": 1},
        )

    def test_transpose(self):
        x = util.AnnotatedArray([[0, 1], [2, 3]], a=(2, 3)).T
        assert (x == [[0, 2], [1, 3]]).all() and x.a == (3, 2)

    def test_any_empty(self):
        assert not util.AnnotatedArray([]).any()

    def test_any(self):
        x = util.AnnotatedArray([[1, 1], [1, 0], [0, 0]], a=(1, 2)).any(
            1, keepdims=True
        )
        assert (x == [[True], [True], [False]]).all() and x.a == (1, 2)

    def test_all_empty(self):
        assert util.AnnotatedArray([]).all()

    def test_all(self):
        x = util.AnnotatedArray([[1, 1], [1, 0], [0, 0]], a=(1, 2)).all(1)
        assert (x == [True, False, False]).all() and x.ann == ({"a": 1},)

    def test_max(self):
        assert util.AnnotatedArray([0, 1, 2], a=(1,)).max() == 2

    def test_min(self):
        assert util.AnnotatedArray([0, 1, 2], a=(1,)).min() == 0

    def test_sum(self):
        assert np.sum(util.AnnotatedArray([1, 2]), initial=-1) == 2

    def test_prod(self):
        assert np.prod(util.AnnotatedArray([1, 2]), initial=-1) == -2

    def test_cumsum(self):
        x = util.AnnotatedArray([[1], [0], [2]], a=(1, 1)).cumsum()
        assert (x == [1, 1, 3]).all() and x.ann == ({"a": 1},)

    def test_cumsum_axis(self):
        x = util.AnnotatedArray([[1, 1], [1, 0], [0, 0]], a=(1, 2)).cumsum(1)
        assert (x == [[1, 2], [1, 1], [0, 0]]).all() and x.a == (1, 2)

    def test_cumprod(self):
        x = util.AnnotatedArray([1, -1, 2], a=(1,)).cumprod()
        assert (x == [1, -1, -2]).all() and x.ann == ({"a": 1},)

    def test_cumprod_axis(self):
        x = util.AnnotatedArray([[1, 1], [1, 0], [0, 0]], a=(1, 2)).cumprod(1)
        assert (x == [[1, 1], [1, 0], [0, 0]]).all() and x.a == (1, 2)

    def test_trace(self):
        x = util.AnnotatedArray([[[1, 2, 3], [4, 5, 6]]], a=(1, 2, 3)).trace(1, 1, 2)
        assert x == 8 and x.ann.as_dict["a"] == (1,)

    def test_imag(self):
        x = util.AnnotatedArray([1 + 2j, 3 + 4j], a=(1,)).imag
        assert (x == [2, 4]).all() and x.ann.as_dict["a"] == (1,)

    def test_real(self):
        x = util.AnnotatedArray([1 + 2j, 3 + 4j], a=(1,)).real
        assert (x == [1, 3]).all() and x.ann.as_dict["a"] == (1,)

    def test_conjugate(self):
        x = util.AnnotatedArray([1 + 2j, 3 + 4j], a=(1,)).conjugate()
        assert (x == [1 - 2j, 3 - 4j]).all() and x.ann.as_dict["a"] == (1,)

    def test_diagonal(self):
        x = util.AnnotatedArray([[[1, 2, 3], [4, 5, 6]]], a=(1, 2, 3)).diagonal(1, 0, 2)
        assert (x == [[2], [5]]).all() and x.ann.as_dict["a"] == (2, None)


class TestImplements:
    def test_solve(self):
        m = util.AnnotatedArray([[1, 2], [3, -1]], a=(3, 1), b=(1,))
        b = util.AnnotatedArray([7, -7], a=(3,))
        a = np.linalg.solve(m, b)
        assert np.all(np.abs(a - [-1, 4]) < 1e-14) and a.ann == ({"a": 1, "b": 1},)

    def test_solve_multiple(self):
        m = util.AnnotatedArray([[1, 2], [3, -1]], a=(3, 1), b=(1,))
        b = util.AnnotatedArray([[7, 3], [-7, 2]], a=(3, 2), b=(3, None))
        a = np.linalg.solve(m, b)
        assert np.all(np.abs(a - [[-1, 1], [4, 1]]) < 1e-14) and a.ann == (
            {"a": 1, "b": 1},
            {"a": 2},
        )

    def test_lstsq(self):
        m = util.AnnotatedArray([[1, 2], [3, -1]], a=(3, 1), b=(1,))
        b = util.AnnotatedArray([7, -7], a=(3,))
        a = np.linalg.lstsq(m, b, None)[0]
        assert (np.abs(a - [-1, 4]) < 1e-14).all() and a.ann == ({"a": 1, "b": 1},)

    def test_lstsq_multiple(self):
        m = util.AnnotatedArray([[1, 2], [3, -1]], a=(3, 1), b=(1,))
        b = util.AnnotatedArray([[7, 3], [-7, 2]], a=(3, 2), b=(3, None))
        a = np.linalg.lstsq(m, b, None)[0]
        assert (np.abs(a - [[-1, 1], [4, 1]]) < 1e-14).all() and a.ann == (
            {"a": 1, "b": 1},
            {"a": 2},
        )

    def test_svd(self):
        m = util.AnnotatedArray([[5, 2], [2, 2]], a=(3, 1))
        u, s, v = np.linalg.svd(m)
        assert (
            (np.abs(s - [6, 1]) < 1e-14).all() and u.a == (3, None) and v.a == (None, 1)
        )

    def test_svd_no_uv(self):
        m = util.AnnotatedArray([[5, 2], [2, 2]], a=(3, 1))
        s = np.linalg.svd(m, compute_uv=False)
        assert (np.abs(s - [6, 1]) < 1e-14).all()

    def test_diag(self):
        x = np.diag(util.AnnotatedArray([[0, 1], [2, 3]], a=(3, 3)))
        assert (x == [0, 3]).all() and x.a == 3

    def test_diag_create(self):
        x = np.diag(util.AnnotatedArray([1], a=(2,)), -1)
        assert (x == [[0, 0], [1, 0]]).all() and x.a == 2

    def test_tril(self):
        x = np.tril(util.AnnotatedArray([[1, 2], [3, 4]], a=(5, 6)))
        assert (x == [[1, 0], [3, 4]]).all() and x.a == (5, 6)

    def test_triu(self):
        x = np.triu(util.AnnotatedArray([[1, 2], [3, 4]], a=(5, 6)), 1)
        assert (x == [[0, 2], [0, 0]]).all() and x.a == (5, 6)

    def test_zeros_like(self):
        x = np.zeros_like(util.AnnotatedArray([1, 2], a=1))
        assert (x == [0, 0]).all() and x.dtype == int and isinstance(x, np.ndarray)

    def test_ones_like(self):
        x = np.ones_like(util.AnnotatedArray([1, 2], a=1))
        assert (x == [1, 1]).all() and x.dtype == int and isinstance(x, np.ndarray)



import itertools

import numpy as np
import scipy.integrate

import treams.special as sc


def _j_real(l0, l1, m, r, dr, pol0, pol1, ks, k, zs, z):
    def fun(theta):
        return np.real(
            np.sin(theta)
            * ((2 * pol0 - 1) * z + (2 * pol1 - 1) * zs)
            * np.dot(
                [r(theta), -dr(theta), 0],
                np.cross(
                    sc.vsw_rA(l0, -m, ks[pol0] * r(theta), theta, 0, pol0),
                    sc.vsw_A(l1, m, k[pol1] * r(theta), theta, 0, pol1),
                ),
            )
        )

    return fun


def _j_imag(l0, l1, m, r, dr, pol0, pol1, ks, k, zs, z):
    def fun(theta):
        return np.imag(
            np.sin(theta)
            * ((2 * pol0 - 1) * z + (2 * pol1 - 1) * zs)
            * np.dot(
                [r(theta), -dr(theta), 0],
                np.cross(
                    sc.vsw_rA(l0, -m, ks[pol0] * r(theta), theta, 0, pol0),
                    sc.vsw_A(l1, m, k[pol1] * r(theta), theta, 0, pol1),
                ),
            )
        )

    return fun


def _rj_real(l0, l1, m, r, dr, pol0, pol1, ks, k, zs, z):
    def fun(theta):
        return np.real(
            np.sin(theta)
            * ((2 * pol0 - 1) * z + (2 * pol1 - 1) * zs)
            * np.dot(
                [r(theta), -dr(theta), 0],
                np.cross(
                    sc.vsw_rA(l0, -m, ks[pol0] * r(theta), theta, 0, pol0),
                    sc.vsw_rA(l1, m, k[pol1] * r(theta), theta, 0, pol1),
                ),
            )
        )

    return fun


def _rj_imag(l0, l1, m, r, dr, pol0, pol1, ks, k, zs, z):
    def fun(theta):
        return np.imag(
            np.sin(theta)
            * ((2 * pol0 - 1) * z + (2 * pol1 - 1) * zs)
            * np.dot(
                [r(theta), -dr(theta), 0],
                np.cross(
                    sc.vsw_rA(l0, -m, ks[pol0] * r(theta), theta, 0, pol0),
                    sc.vsw_rA(l1, m, k[pol1] * r(theta), theta, 0, pol1),
                ),
            )
        )

    return fun


def qmat(r, dr, ks, zs, out, in_=None, singular=True):
    l_out, m_out, pol_out = out
    in_ = out if in_ is None else in_
    l_in, m_in, pol_in = in_
    if singular:
        fr = _j_real
        fi = _j_imag
    else:
        fr = _rj_real
        fi = _rj_imag
    size_out = len(l_out)
    size_in = len(l_in)
    res = np.zeros((size_out, size_in), complex)
    for i, j in itertools.product(range(size_out), range(size_in)):
        if m_out[i] != m_in[j]:
            continue
        res[i, j] = (
            scipy.integrate.quad(
                fr(l_out[i], l_in[j], m_out[i], r, dr, pol_out[i], pol_in[j], *ks, *zs),
                0,
                np.pi,
            )[0]
            + 1j
            * scipy.integrate.quad(
                fi(l_out[i], l_in[j], m_out[i], r, dr, pol_out[i], pol_in[j], *ks, *zs),
                0,
                np.pi,
            )[0]
        )
    return res



import matplotlib.pyplot as plt
import numpy as np

import treams

k0s = 2 * np.pi * np.linspace(1 / 1000, 1 / 300, 50)
materials = [(), (12.4 + 1j, 1 + 0.1j, 0.5 + 0.05j), (2, 2)]
thickness = 50
tr = np.zeros((2, len(k0s), 2))
for i, k0 in enumerate(k0s):
    pwb = treams.PlaneWaveBasisByComp.default([0, 0.5 * k0])
    slab = treams.SMatrices.slab(thickness, pwb, k0, materials)
    tr[0, i, :] = slab.tr([1, 0])
    tr[1, i, :] = slab.tr([0, 1])

fig, ax = plt.subplots()
ax.set_xlabel("Frequency (THz)")
ax.plot(299792.458 * k0s / (2 * np.pi), tr[0, :, 0])
ax.plot(299792.458 * k0s / (2 * np.pi), tr[0, :, 1])
ax.plot(299792.458 * k0s / (2 * np.pi), tr[1, :, 0], c="C0", linestyle="--")
ax.plot(299792.458 * k0s / (2 * np.pi), tr[1, :, 1], c="C1", linestyle="--")
ax.legend(["$T_+$", "$R_+$", "$T_-$", "$R_-$"])
fig.show()



"""Basis sets and core array functionalities."""

import abc
from collections import namedtuple

import numpy as np

import treams._operators as op
import treams.lattice as la
from treams import util
from treams._lattice import Lattice, WaveVector
from treams._material import Material


class BasisSet(util.OrderedSet, metaclass=abc.ABCMeta):
    """Basis set base class.

    It is the base class for all basis sets used. They are expected to be an ordered
    sequence of the modes, that are included in a expansion. Basis sets are expected to
    be immutable.
    """

    _names = ()
    """Names of the relevant parameters"""

    def __repr__(self):
        """String representation.

        Automatically generated when the attribute ``_names`` is defined.

        Returns:
            str
        """
        string = ",\n    ".join(f"{name}={i}" for name, i in zip(self._names, self[()]))
        return f"{self.__class__.__name__}(\n    {string},\n)"

    def __len__(self):
        """Number of modes."""
        return len(self.pol)

    @classmethod
    @abc.abstractmethod
    def default(cls, *args, **kwargs):
        """Construct a default basis from parameters.

        Construct a basis set in a default order by giving few parameters.
        """
        raise NotImplementedError


class SphericalWaveBasis(BasisSet):
    r"""Basis of spherical waves.

    Functions of the spherical wave basis are defined by their angular momentum ``l``,
    its projection onto the z-axis ``m``, and the polarization ``pol``. If the basis
    is defined with respect to a single origin it is referred to as "global", if it
    contains multiple origins it is referred to as "local". In a local basis an
    additional position index ``pidx`` is used to link the modes to one of the
    specified ``positions``.

    For spherical waves there exist multiple :ref:`params:Polarizations` and they
    can be separated into incident and scattered fields. Depending on these combinations
    the basis modes refer to one of the functions :func:`~treams.special.vsw_A`,
    :func:`~treams.special.vsw_rA`, :func:`~treams.special.vsw_M`,
    :func:`~treams.special.vsw_rM`, :func:`~treams.special.vsw_N`, or
    :func:`~treams.special.vsw_rN`.

    Args:
        modes (array-like): A tuple containing a list for each of ``l``, ``m``, and
            ``pol`` or ``pidx``, ``l``, ``m``, and``pol``.
        positions (array-like, optional): The positions of the origins for the specified
            modes. Defaults to ``[[0, 0, 0]]``.

    Attributes:
        pidx (array-like): Integer referring to a row in :attr:`positions`.
        l (array-like): Angular momentum as an integer :math:`l > 0`
        m (array-like): Angular momentum projection onto the z-axis, it is an integer
            with :math:`m \leq |l|`
        pol (array-like): Polarization, see also :ref:`params:Polarizations`.
    """

    _names = ("pidx", "l", "m", "pol")

    def __init__(self, modes, positions=None):
        """Initalization."""
        tmp = []
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            pidx = []
            l = []  # noqa: E741
            m = []
            pol = []
        elif len(modes[0]) == 3:
            l, m, pol = (*zip(*modes),)
            pidx = np.zeros_like(l)
        elif len(modes[0]) == 4:
            pidx, l, m, pol = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        if positions is None:
            positions = np.zeros((1, 3))
        positions = np.array(positions, float)
        if positions.ndim == 1:
            positions = positions[None, :]
        if positions.ndim != 2 or positions.shape[1] != 3:
            raise ValueError(f"invalid shape of positions {positions.shape}")

        self.pidx, self.l, self.m, self.pol = [
            np.array(i, int) for i in (pidx, l, m, pol)
        ]
        for i, j in ((self.pidx, pidx), (self.l, l), (self.m, m), (self.pol, pol)):
            i.flags.writeable = False
            if np.any(i != j):
                raise ValueError("parameters must be integer")
        if np.any(self.l < 1):
            raise ValueError("'l' must be a strictly positive integer")
        if np.any(self.l < np.abs(self.m)):
            raise ValueError("'|m|' cannot be larger than 'l'")
        if np.any(self.pol > 1) or np.any(self.pol < 0):
            raise ValueError("polarization must be '0' or '1'")
        if np.any(self.pidx >= len(positions)):
            raise ValueError("undefined position is indexed")

        self._positions = positions
        self._positions.flags.writeable = False
        self.lattice = self.kpar = None

    @property
    def positions(self):
        """Positions of the modes' origins.

        The positions are an immutable (N, 3)-array. Each row corresponds to a point in
        the three-dimensional Cartesian space.
        """
        return self._positions

    def __repr__(self):
        """String representation."""
        positions = "positions=" + str(self.positions).replace("\n", ",")
        return f"{super().__repr__()[:-1]}    {positions},\n)"

    @property
    def isglobal(self):
        """Basis is defined with respect to a single (global) origin.

        Returns:
            bool
        """
        return len(self) == 0 or np.all(self.pidx == self.pidx[0])

    def __getattr__(self, key):
        dct = {"l": "l", "m": "m", "p": "pidx", "s": "pol"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function allows index into the basis set by an integer, a slice, a sequence
        of integers or bools, an ellipsis, or an empty tuple. All of them except the
        integer and the empty tuple results in another basis set being returned. In case
        of the two exceptions a tuple is returned.

        Alternatively, the string "plmp", "lmp", or "lm" can be used to access only a
        subset of :attr:`pidx`, :attr:`l`, :attr:`m`, and :attr:`pol`.
        """
        res = self.pidx[idx], self.l[idx], self.m[idx], self.pol[idx]
        if isinstance(idx, (int, np.integer)) or (
            isinstance(idx, tuple) and len(idx) == 0
        ):
            return res
        return type(self)(zip(*res), self.positions)

    def __eq__(self, other):
        """Compare basis sets.

        Basis sets are considered equal, when they have the same modes in the same order
        and the specified origin :attr:`positions` are equal.
        """
        try:
            return self is other or (
                np.array_equal(self.pidx, other.pidx)
                and np.array_equal(self.l, other.l)
                and np.array_equal(self.m, other.m)
                and np.array_equal(self.pol, other.pol)
                and np.array_equal(self.positions, other.positions)
            )
        except AttributeError:
            return False

    @classmethod
    def default(cls, lmax, nmax=1, positions=None):
        """Default basis for the given maximal multipolar order.

        The default order contains separate blocks for each position index which are in
        ascending order. Within each block the modes are sorted by angular momentum
        :math:`l`, with the lowest angular momentum coming first. For each angular
        momentum its z-projection is in ascending order from :math:`m = -l` to
        :math:`m = l`. Finally, the polarization index is the fastest changing index
        which iterates between 1 and 0.

        Example:
            >>> SphericalWaveBasis.default(2)
            SphericalWaveBasis(
                pidx=[0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0],
                l=[1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2],
                m=[-1 -1  0  0  1  1 -2 -2 -1 -1  0  0  1  1  2  2],
                pol=[1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0],
                positions=[[0. 0. 0.]],
            )
            >>> SphericalWaveBasis.default(1, 2, [[0, 0, 1.], [0, 0, -1.]])
            SphericalWaveBasis(
                pidx=[0 0 0 0 0 0 1 1 1 1 1 1],
                l=[1 1 1 1 1 1 1 1 1 1 1 1],
                m=[-1 -1  0  0  1  1 -1 -1  0  0  1  1],
                pol=[1 0 1 0 1 0 1 0 1 0 1 0],
                positions=[[ 0.  0.  1.], [ 0.  0. -1.]],
            )

        Args:
            lmax (int): Maximal multipolar order.
            nmax (int, optional): Number of positions, defaults to 1.
            positions (array-like, optional): Positions of the origins.
        """
        modes = [
            [n, l, m, p]
            for n in range(0, nmax)
            for l in range(1, lmax + 1)  # noqa: E741
            for m in range(-l, l + 1)
            for p in range(1, -1, -1)
        ]
        return cls(modes, positions=positions)

    @classmethod
    def ebcm(cls, lmax, nmax=1, mmax=-1, positions=None):
        """Order of modes suited for ECBM.

        In comparison to :meth:`default` this order prioritises blocks of the
        z-projection of the angular momentum :math:`m` over the angular momentum
        :math:`l`. This is useful to get block-diagonal matrices for the extended
        boundary condition method (EBCM).

        Example:
            >>> SphericalWaveBasis.ebcm(2)
            SphericalWaveBasis(
                pidx=[0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0],
                l=[2 2 1 1 2 2 1 1 2 2 1 1 2 2 2 2],
                m=[-2 -2 -1 -1 -1 -1  0  0  0  0  1  1  1  1  2  2],
                pol=[1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0],
                positions=[[0. 0. 0.]],
            )

        Args:
            lmax (int): Maximal multipolar order.
            nmax (int, optional): Number of particles, defaults to 1.
            mmax (int, optional): Maximal value of `|m|`. If ommitted or set to -1 it
                is taken equal to `lmax`.
            positions (array-like, optional): Positions of the origins.

        Returns:
            tuple
        """
        mmax = lmax if mmax == -1 else mmax
        modes = [
            [n, l, m, p]
            for n in range(0, nmax)
            for m in range(-mmax, mmax + 1)
            for l in range(max(abs(m), 1), lmax + 1)  # noqa: E741
            for p in range(1, -1, -1)
        ]
        return cls(modes, positions=positions)

    @staticmethod
    def defaultlmax(dim, nmax=1):
        """Calculate the default mode order for a given length.

        Given the dimension of the T-matrix return the estimated maximal value of `l`.
        This is the inverse of :meth:`defaultdim`. A value of zero is allowed for empty
        T-matrices.

        Example:
            >>> SphericalWaveBasis.defaultlmax(len(SphericalWaveBasis.default(3)))
            3

        Args:
            dim (int): Dimension of the T-matrix, respectively number of modes.
            nmax (int, optional): Number of particles, defaults to 1.

        Returns:
            int
        """
        res = np.sqrt(1 + dim * 0.5 / nmax) - 1
        res_int = int(np.rint(res))
        if np.abs(res - res_int) > 1e-8 * np.maximum(np.abs(res), np.abs(res_int)):
            raise ValueError("cannot estimate the default lmax")
        return res_int

    @staticmethod
    def defaultdim(lmax, nmax=1):
        """Default number of modes for a given mulipolar order.

        Given the maximal value of `l` return the size of the corresponding T-matrix.
        This is the inverse of :meth:`defaultlmax`. A value of zero is allowed.

        Args:
            lmax (int): Maximal multipolar order
            nmax (int, optional): Number of particles, defaults to `1`

        Returns:
            int
        """
        # zero is allowed and won't give an error
        if lmax < 0 or nmax < 0:
            raise ValueError("maximal order must be positive")
        return 2 * lmax * (lmax + 2) * nmax

    @classmethod
    def _from_iterable(cls, it, positions=None):
        if isinstance(cls, SphericalWaveBasis):
            positions = cls.positions if positions is None else positions
            cls = type(cls)
        obj = cls(it, positions=positions)
        return obj


class CylindricalWaveBasis(BasisSet):
    r"""Basis of cylindrical waves.

    Functions of the cylindrical wave basis are defined by the z-components of the wave
    vector ``kz`` and the angular momentum  ``m`` as well as the polarization ``pol``.
    If the basis is defined with respect to a single origin it is referred to as
    "global", if it contains multiple origins it is referred to as "local". In a local
    basis an additional position index ``pidx`` is used to link the modes to one of the
    specified ``positions``.

    For cylindrical waves there exist multiple :ref:`params:Polarizations` and
    they can be separated into incident and scattered fields. Depending on these
    combinations the basis modes refer to one of the functions
    :func:`~treams.special.vcw_A`, :func:`~treams.special.vcw_rA`,
    :func:`~treams.special.vcw_M`, :func:`~treams.special.vcw_rM`,
    :func:`~treams.special.vcw_N`, or :func:`~treams.special.vcw_rN`.

    Args:
        modes (array-like): A tuple containing a list for each of ``kz``, ``m``, and
            ``pol`` or ``pidx``, ``kz``, ``m``, and``pol``.
        positions (array-like, optional): The positions of the origins for the specified
            modes. Defaults to ``[[0, 0, 0]]``.

    Attributes:
        pidx (array-like): Integer referring to a row in :attr:`positions`.
        kz (array-like): Real valued z-component of the wave vector.
        m (array-like): Integer angular momentum projection onto the z-axis.
        pol (array-like): Polarization, see also :ref:`params:Polarizations`.
    """

    _names = ("pidx", "kz", "m", "pol")

    def __init__(self, modes, positions=None):
        """Initalization."""
        tmp = []
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            pidx = []
            kz = []
            m = []
            pol = []
        elif len(modes[0]) == 3:
            kz, m, pol = (*zip(*modes),)
            pidx = np.zeros_like(m)
        elif len(modes[0]) == 4:
            pidx, kz, m, pol = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        if positions is None:
            positions = np.zeros((1, 3))
        positions = np.array(positions, float)
        if positions.ndim == 1:
            positions = positions[None, :]
        if positions.ndim != 2 or positions.shape[1] != 3:
            raise ValueError(f"invalid shape of positions {positions.shape}")

        self.pidx, self.m, self.pol = [np.array(i, int) for i in (pidx, m, pol)]
        self.kz = np.array(kz, float)
        self.kz.flags.writeable = False
        for i, j in ((self.pidx, pidx), (self.m, m), (self.pol, pol)):
            i.flags.writeable = False
            if np.any(i != j):
                raise ValueError("parameters must be integer")
        if np.any(self.pol > 1) or np.any(self.pol < 0):
            raise ValueError("polarization must be '0' or '1'")
        if np.any(self.pidx >= len(positions)):
            raise ValueError("undefined position is indexed")

        self._positions = positions
        self._positions.flags.writeable = False

        self.lattice = self.kpar = None
        if len(self.kz) > 0 and np.all(self.kz == self.kz[0]):
            self.kpar = WaveVector(self.kz[0])

    @property
    def positions(self):
        """Positions of the modes' origins.

        The positions are an immutable (N, 3)-array. Each row corresponds to a point in
        the three-dimensional Cartesian space.
        """
        return self._positions

    def __repr__(self):
        """String representation."""
        positions = "positions=" + str(self.positions).replace("\n", ",")
        return f"{super().__repr__()[:-1]}    {positions},\n)"

    @property
    def isglobal(self):
        """Basis is defined with respect to a single (global) origin.

        Returns:
            bool
        """
        return len(self) == 0 or np.all(self.pidx == self.pidx[0])

    def __getattr__(self, key):
        dct = {"z": "kz", "m": "m", "p": "pidx", "s": "pol"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function allows index into the basis set by an integer, a slice, a sequence
        of integers or bools, an ellipsis, or an empty tuple. All of them except the
        integer and the empty tuple results in another basis set being returned. In case
        of the two exceptions a tuple is returned.

        Alternatively, the string "pkzmp", "kzmp", or "kzm" can be used to access only a
        subset of :attr:`pidx`, :attr:`kz`, :attr:`m`, and :attr:`pol`.
        """
        res = self.pidx[idx], self.kz[idx], self.m[idx], self.pol[idx]
        if isinstance(idx, (int, np.integer)) or (isinstance(idx, tuple) and idx == ()):
            return res
        return type(self)(zip(*res), self.positions)

    def __eq__(self, other):
        """Compare basis sets.

        Basis sets are considered equal, when they have the same modes in the same order
        and the specified origin :attr:`positions` are equal.
        """
        try:
            return self is other or (
                np.array_equal(self.pidx, other.pidx)
                and np.array_equal(self.kz, other.kz)
                and np.array_equal(self.m, other.m)
                and np.array_equal(self.pol, other.pol)
                and np.array_equal(self.positions, other.positions)
            )
        except AttributeError:
            return False

    @classmethod
    def default(cls, kzs, mmax, nmax=1, positions=None):
        """Default basis for the given z-components of wave vector and angular momentum.

        The default order contains separate blocks for each position index which are in
        ascending order. Within each block the modes are sorted by the z-component of
        the wave vector :math:`k_z`. For each of those values the z-projection of the
        angular momentum is placed in ascending order. Finally, the polarization index
        is the fastest changing index which iterates between 1 and 0.

        Example:
            >>> CylindricalWaveBasis.default([-0.5, 0.5], 1)
            CylindricalWaveBasis(
                pidx=[0 0 0 0 0 0 0 0 0 0 0 0],
                kz=[-0.5 -0.5 -0.5 -0.5 -0.5 -0.5  0.5  0.5  0.5  0.5  0.5  0.5],
                m=[-1 -1  0  0  1  1 -1 -1  0  0  1  1],
                pol=[1 0 1 0 1 0 1 0 1 0 1 0],
                positions=[[0. 0. 0.]],
            )
            >>> CylindricalWaveBasis.default([0], 1, 2, [[1., 0, 0], [-1., 0, 0]])
            CylindricalWaveBasis(
                pidx=[0 0 0 0 0 0 1 1 1 1 1 1],
                kz=[0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.],
                m=[-1 -1  0  0  1  1 -1 -1  0  0  1  1],
                pol=[1 0 1 0 1 0 1 0 1 0 1 0],
                positions=[[ 1.  0.  0.], [-1.  0.  0.]],
            )

        Args:
            kzs (array-like, float): Maximal multipolar order.
            mmax (int): Maximal value of the angular momentum z-component.
            nmax (int, optional): Number of positions, defaults to 1.
            positions (array-like, optional): Positions of the origins.
        """
        kzs = np.atleast_1d(kzs)
        if kzs.ndim > 1:
            raise ValueError(f"kzs has dimension larger than one: '{kzs.ndim}'")
        modes = [
            [n, kz, m, p]
            for n in range(nmax)
            for kz in kzs
            for m in range(-mmax, mmax + 1)
            for p in range(1, -1, -1)
        ]
        return cls(modes, positions=positions)

    @classmethod
    def diffr_orders(cls, kz, mmax, lattice, bmax, nmax=1, positions=None):
        """Create a basis set for a system periodic in the z-direction.

        Example:
           >>> CylindricalWaveBasis.diffr_orders(0.1, 1, 2 * np.pi, 1)
           CylindricalWaveBasis(
               pidx=[0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0],
               kz=[-0.9 -0.9 -0.9 -0.9 -0.9 -0.9  0.1  0.1  0.1  0.1  0.1  0.1  1.1  1.1
             1.1  1.1  1.1  1.1],
               m=[-1 -1  0  0  1  1 -1 -1  0  0  1  1 -1 -1  0  0  1  1],
               pol=[1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0 1 0],
               positions=[[0. 0. 0.]],
           )

        Args:
            kz (float): Wave vector z-component. Ideally it is in the first Brillouin
                zone (use :func:`misc.firstbrillouin1d`).
            mmax (int): Maximal value for the z-component of the angular momentum.
            lattice (:class:`treams.Lattice` or float): Lattice definition or pitch.
            bmax (float): Maximal change of the z-component of the wave vector. So,
                this defines a maximal momentum transfer from the given value `kz`.
            nmax (int, optional): Number of positions.
            positions (array-like, optional): Positions of the origins.
        """
        lattice = Lattice(lattice)
        lattice_z = Lattice(lattice, "z")
        nkz = np.floor(np.abs(bmax / lattice_z.reciprocal))
        kzs = kz + np.arange(-nkz, nkz + 1) * lattice_z.reciprocal
        res = cls.default(kzs, mmax, nmax, positions=positions)
        res.lattice = lattice
        res.kpar = WaveVector(kz)
        return res

    @classmethod
    def _from_iterable(cls, it, positions=None):
        if isinstance(cls, CylindricalWaveBasis):
            positions = cls.positions if positions is None else positions
            lattice = cls.lattice
            kpar = cls.kpar
            cls = type(cls)
        else:
            lattice = kpar = None
        obj = cls(it, positions=positions)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    @staticmethod
    def defaultmmax(dim, nkz=1, nmax=1):
        """Calculate the default mode order for a given length.

        Given the dimension of the T-matrix return the estimated maximal value of `m`.
        This is the inverse of :meth:`defaultdim`. A value of zero is allowed for empty
        T-matrices.

        Example:
            >>> CylindricalWaveBasis.defaultmmax(len(CylindricalWaveBasis.default([0], 2)), 1)
            2

        Args:
            dim (int): Dimension of the T-matrix, respectively number of modes
            nkz (int, optional): Number of z-components of the wave vector.
            nmax (int, optional): Number of particles, defaults to 1.

        Returns:
            int
        """
        if dim % (2 * nkz * nmax) != 0:
            raise ValueError("cannot estimate the default mmax")
        dim = dim // (2 * nkz * nmax)
        if (dim - 1) % 2 != 0:
            raise ValueError("cannot estimate the default mmax")
        return (dim - 1) // 2

    @staticmethod
    def defaultdim(nkz, mmax, nmax=1):
        """Default number of modes for a given mulipolar order.

        Given the maximal value of `l` return the size of the corresponding T-matrix.
        This is the inverse of :meth:`defaultlmax`. A value of zero is allowed.

        Args:
            nkz (int): Number of z-components of the wave vector.
            mmax (int): Maximal value of the angular momentum's z-component.
            nmax (int, optional): Number of particles, defaults to 1.

        Returns:
            int
        """
        if nkz < 0 or mmax < 0:
            raise ValueError("maximal order must be positive")
        return (2 * mmax + 1) * nkz * 2 * nmax


class PlaneWaveBasis(BasisSet):
    """Plane wave basis parent class."""

    isglobal = True


class PlaneWaveBasisByUnitVector(PlaneWaveBasis):
    """Plane wave basis.

    A plane wave basis is defined by a collection of wave vectors specified by the
    Cartesian wave vector components ``qx``, ``qy``, and ``qz`` normalized to
    :math:`q_x^2 + q_y^2 + q_z^2 = 1` and the polarizations ``pol``.

    For plane waves there exist multiple :ref:`params:Polarizations`, such that
    these modes can refer to either :func:`~treams.special.vpw_A` or
    :func:`~treams.special.vpw_M` and :func:`~treams.special.vpw_N`.

    Args:
        modes (array-like): A tuple containing a list for each of ``qx``, ``qy``,
            ``qz``, and ``pol``.

    Attributes:
        qx (array-like): X-component of the normalized wave vector.
        qy (array-like): Y-component of the normalized wave vector.
        qz (array-like): Z-component of the normalized wave vector.
        pol (array-like): Polarization, see also :ref:`params:Polarizations`.
    """

    _names = ("qx", "qy", "qz", "pol")
    """A plane wave basis is always global."""

    def __init__(self, modes):
        """Initialization."""
        tmp = []
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            qx = []
            qy = []
            qz = []
            pol = []
        elif len(modes[0]) == 4:
            qx, qy, qz, pol = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        qx, qy, qz, pol = (np.array(i) for i in (qx, qy, qz, pol))
        norm = np.emath.sqrt(qx * qx + qy * qy + qz * qz)
        norm[np.abs(norm - 1) < 1e-14] = 1
        qx, qy, qz = (np.true_divide(i, norm) for i in (qx, qy, qz))
        for i in (qx, qy, qz, pol):
            i.flags.writeable = False
            if i.ndim > 1:
                raise ValueError("invalid shape of parameters")
        self.qx, self.qy, self.qz = qx, qy, qz
        self.pol = np.array(pol.real, int)
        if np.any(self.pol != pol):
            raise ValueError("polarizations must be integer")
        if np.any(self.pol > 1) or np.any(self.pol < 0):
            raise ValueError("polarization must be '0' or '1'")
        self.lattice = self.kpar = None

    def __getattr__(self, key):
        dct = {"x": "qx", "y": "qy", "z": "qz", "s": "pol"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function allows index into the basis set by an integer, a slice, a sequence
        of integers or bools, an ellipsis, or an empty tuple. All of them except the
        integer and the empty tuple results in another basis set being returned. In case
        of the two exceptions a tuple is returned.

        Alternatively, the string "xyzp", "xyp", or "zp" can be used to access only a
        subset of :attr:`qx`, :attr:`qy`, :attr:`qz`, and :attr:`pol`.
        """
        res = self.qx[idx], self.qy[idx], self.qz[idx], self.pol[idx]
        if isinstance(idx, (int, np.integer)) or (isinstance(idx, tuple) and idx == ()):
            return res
        return type(self)(zip(*res))

    @classmethod
    def default(cls, kvecs):
        """Default basis from the given wave vectors.

        For each wave vector the two polarizations 1 and 0 are taken.

        Example:
            >>> PlaneWaveBasisByUnitVector.default([[0, 0, 5], [0, 3, 4]])
            PlaneWaveBasisByUnitVector(
                qx=[0. 0. 0. 0.],
                qy=[0.  0.  0.6 0.6],
                qz=[1.  1.  0.8 0.8],
                pol=[1 0 1 0],
            )

        Args:
            kvecs (array-like): Wave vectors in Cartesian coordinates.
        """
        kvecs = np.atleast_2d(kvecs)
        modes = np.empty((2 * kvecs.shape[0], 4), kvecs.dtype)
        modes[::2, :3] = kvecs
        modes[1::2, :3] = kvecs
        modes[::2, 3] = 1
        modes[1::2, 3] = 0
        return cls(modes)

    @classmethod
    def _from_iterable(cls, it):
        if isinstance(cls, PlaneWaveBasisByUnitVector):
            lattice = cls.lattice
            kpar = cls.kpar
            cls = type(cls)
        else:
            lattice = kpar = None
        obj = cls(it)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    def __eq__(self, other):
        """Compare basis sets.

        Basis sets are considered equal, when they have the same modes in the same
        order.
        """
        try:
            return self is other or (
                np.array_equal(self.qx, other.qx)
                and np.array_equal(self.qy, other.qy)
                and np.array_equal(self.qz, other.qz)
                and np.array_equal(self.pol, other.pol)
            )
        except AttributeError:
            return False

    def bycomp(self, k0, alignment="xy", material=Material()):
        """Create a :class:`PlaneWaveBasisByComp`.

        The plane wave basis is changed to a partial basis, where only two (real-valued)
        wave vector components are defined the third component is then inferred from the
        dispersion relation, which depends on the wave number and material, and a
        ``modetype`` that specifies the sign of the third component.

        Args:
            alignment (str, optional): Wave vector components that are part of the
                partial basis. Defaults to "xy", other permitted values are "yz" and
                "zx".
            k0 (float, optional): Wave number. If given, it is checked that the current
                basis fulfils the dispersion relation.
            material (:class:`~treams.Material` or tuple): Material definition. Defaults
                to vacuum/air.
        """
        ks = material.ks(k0)[self.pol]
        if alignment in ("xy", "yz", "zx"):
            kpars = [ks * getattr(self, "q" + s) for s in alignment]
        else:
            raise ValueError(f"invalid alignment '{alignment}'")
        obj = PlaneWaveBasisByComp(zip(*kpars, self.pol), alignment)
        obj.lattice = self.lattice
        obj.kpar = self.kpar
        return obj

    def kvecs(self, k0, material=Material(), modetype=None):
        """Wave vectors.

        Args:
            k0 (float): Wave number.
            material (:class:`~treams.Material` or tuple, optional): Material
                definition. Defaults to vacuum/air.
            modetype (optional): Currently unused for this class.
        """
        # TODO: check kz depending on modetype (alignment?)
        ks = Material(material).ks(k0)[self.pol]
        return ks * self.qx, ks * self.qy, ks * self.qz

    def permute(self, n=1):
        n = n % 3
        lattice = None if self.lattice is None else self.lattice.permute(n)
        kpar = None if self.kpar is None else self.kpar.permute(n)
        qx, qy, qz = self.qx, self.qy, self.qz
        for _ in range(n):
            qx, qy, qz = qz, qx, qy
        obj = type(self)(zip(*(qx, qy, qz, self.pol)))
        obj.lattice = lattice
        obj.kpar = kpar
        return obj


class PlaneWaveBasisByComp(PlaneWaveBasis):
    """Partial plane wave basis.

    A partial plane wave basis is defined by two wave vector components out of all three
    Cartesian wave vector components ``kx``, ``ky``, and ``kz`` and the polarizations
    ``pol``. Which two components are given is specified in the :attr:`alignment`. This
    basis is mostly used for stratified media that are periodic or uniform in the two
    alignment directions, such that the given wave vector components correspond to the
    diffraction orders.

    For plane waves there exist multiple :ref:`params:Polarizations`, such that
    these modes can refer to either :func:`~treams.special.vpw_A` or
    :func:`~treams.special.vpw_M` and :func:`~treams.special.vpw_N`.

    Args:
        modes (array-like): A tuple containing a list for each of ``k1``, ``k2``, and
            ``pol``.
        alignment (str, optional): Definition which wave vector components are given.
            Defaults to "xy", other possible values are "yz" and "zx".

    Attributes:
        alignment (str): Alignment of the partial basis.
        pol (array-like): Polarization, see also :ref:`params:Polarizations`.
    """

    def __init__(self, modes, alignment="xy"):
        """Initialization."""
        if isinstance(modes, np.ndarray):
            modes = modes.tolist()
        tmp = []
        for m in modes:
            if m not in tmp:
                tmp.append(m)
        modes = tmp
        if len(modes) == 0:
            kx = []
            ky = []
            pol = []
        elif len(modes[0]) == 3:
            kx, ky, pol = (*zip(*modes),)
        else:
            raise ValueError("invalid shape of modes")

        self._kx, self._ky = [np.real(i) for i in (kx, ky)]
        self.pol = np.array(np.real(pol), int)
        for i, j in [(self._kx, kx), (self._ky, ky), (self.pol, pol)]:
            i.flags.writeable = False
            if i.ndim > 1:
                raise ValueError("invalid shape of parameters")
            if np.any(i != j):
                raise ValueError("invalid value for parameter, must be real")
        if np.any(self.pol > 1) or np.any(self.pol < 0):
            raise ValueError("polarization must be '0' or '1'")

        if alignment in ("xy", "yz", "zx"):
            self._names = (*("k" + i for i in alignment),) + ("pol",)
        else:
            raise ValueError(f"invalid alignment '{alignment}'")

        self.alignment = alignment
        self.lattice = self.kpar = None

    def permute(self, n=1):
        n = n % 3
        lattice = None if self.lattice is None else self.lattice.permute(n)
        kpar = None if self.kpar is None else self.kpar.permute(n)
        alignments = {"xy": "yz", "yz": "zx", "zx": "xy"}
        alignment = self.alignment
        for _ in range(n):
            alignment = alignments[alignment]
        obj = self._from_iterable(self, alignment=alignment)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    @property
    def kx(self):
        """X-components of the wave vector.

        If the components are not specified `None` is returned.
        """
        if self.alignment == "xy":
            return self._kx
        if self.alignment == "zx":
            return self._ky
        return None

    @property
    def ky(self):
        """Y-components of the wave vector.

        If the components are not specified `None` is returned.
        """
        if self.alignment == "xy":
            return self._ky
        if self.alignment == "yz":
            return self._kx
        return None

    @property
    def kz(self):
        """Z-components of the wave vector.

        If the components are not specified `None` is returned.
        """
        if self.alignment == "yz":
            return self._ky
        if self.alignment == "zx":
            return self._kx
        return None

    def __getattr__(self, key):
        dct = {"x": "kx", "y": "ky", "z": "kz", "s": "pol"}
        try:
            return tuple(getattr(self, dct[k]) for k in key)
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            ) from None

    def __getitem__(self, idx):
        """Get a subset of the basis.

        This function allows index into the basis set by an integer, a slice, a sequence
        of integers or bools, an ellipsis, or an empty tuple. All of them except the
        integer and the empty tuple results in another basis set being returned. In case
        of the two exceptions a tuple is returned.
        """
        res = self._kx[idx], self._ky[idx], self.pol[idx]
        if isinstance(idx, (int, np.integer)) or (isinstance(idx, tuple) and idx == ()):
            return res
        return type(self)(zip(*res))

    @classmethod
    def default(cls, kpars, alignment="xy"):
        """Default basis from the given wave vectors.

        For each wave vector the two polarizations 1 and 0 are taken.

        Example:
            >>> PlaneWaveBasisByComp.default([[0, 0], [0, 3]])
            PlaneWaveBasisByComp(
                kx=[0. 0. 0. 0.],
                ky=[0. 0. 3. 3.],
                pol=[1 0 1 0],
            )

        Args:
            kpars (array-like): Wave vector components in Cartesian coordinates.
            alignment (str, optional): Definition which wave vector components are
                given. Defaults to "xy", other possible values are "yz" and "zx".

        """
        kpars = np.atleast_2d(kpars)
        modes = np.empty((2 * kpars.shape[0], 3), float)
        modes[::2, :2] = kpars
        modes[1::2, :2] = kpars
        modes[::2, 2] = 1
        modes[1::2, 2] = 0
        return cls(modes, alignment=alignment)

    @classmethod
    def diffr_orders(cls, kpar, lattice, bmax):
        """Create a basis set for a two-dimensional periodic system.

        The reciprocal lattice to the given lattice is taken to consider all diffraction
        orders that lie within the defined maximal radius (in reciprocal space).

        Example:
            >>> PlaneWaveBasisByComp.diffr_orders([0, 0], Lattice.square(2 * np.pi), 1)
            PlaneWaveBasisByComp(
                kx=[ 0.  0.  0.  0.  0.  0.  1.  1. -1. -1.],
                ky=[ 0.  0.  1.  1. -1. -1.  0.  0.  0.  0.],
                pol=[1 0 1 0 1 0 1 0 1 0],
            )

        Args:
            kpar (float): Tangential wave vector components. Ideally they are in the
                first Brillouin zone (use :func:`misc.firstbrillouin2d`).
            lattice (:class:`treams.Lattice` or float): Lattice definition or pitch.
            bmax (float): Maximal change of tangential wave vector components. So,
                this defines a maximal momentum transfer.
        """
        lattice = Lattice(lattice)
        if lattice.dim != 2:
            raise ValueError("invalid lattice dimensions")
        latrec = lattice.reciprocal
        kpars = kpar + la.diffr_orders_circle(latrec, bmax) @ latrec
        obj = cls.default(kpars, alignment=lattice.alignment)
        obj.lattice = lattice
        obj.kpar = WaveVector(kpar, alignment=lattice.alignment)
        return obj

    @classmethod
    def _from_iterable(cls, it, alignment="xy"):
        if isinstance(cls, PlaneWaveBasisByComp):
            alignment = cls.alignment if alignment is None else alignment
            lattice = cls.lattice
            kpar = cls.kpar
            cls = type(cls)
        else:
            lattice = kpar = None
        obj = cls(it, alignment)
        obj.lattice = lattice
        obj.kpar = kpar
        return obj

    def __eq__(self, other):
        """Compare basis sets.

        Basis sets are considered equal, when they have the same modes in the same
        order.
        """
        try:
            skx, sky, skz = self.kx, self.ky, self.kz
            okx, oky, okz = other.kx, other.ky, other.kz
            return self is other or (
                (np.array_equal(skx, okx) or (skx is None and okx is None))
                and (np.array_equal(sky, oky) or (sky is None and oky is None))
                and (np.array_equal(skz, okz) or (skz is None and okz is None))
                and np.array_equal(self.pol, other.pol)
            )
        except AttributeError:
            return False

    def byunitvector(self, k0, material=Material(), modetype="up"):
        """Create a complete basis :class:`PlaneWaveBasis`.

        A plane wave basis is considered complete, when all three Cartesian components
        and the polarization is defined for each mode. So, the specified wave number,
        material, and modetype is taken to calculate the third Cartesian wave vector.
        The modetype "up" ("down") is for waves propagating in the positive (negative)
        direction with respect to the Cartesian axis that is orthogonal to the
        alignment.

        Args:
            k0 (float): Wave number.
            material (:class:`~treams.Material` or tuple, optional): Material
                definition. Defaults to vacuum/air.
            modetype (str, optional): Propagation direction. Defaults to "up".
        """
        if modetype not in ("up", "down"):
            raise ValueError("modetype not recognized")
        material = Material(material)
        kx = self._kx
        ky = self._ky
        kz = material.kzs(k0, kx, ky, self.pol) * (2 * (modetype == "up") - 1)
        if self.alignment == "yz":
            kx, ky, kz = kz, kx, ky
        elif self.alignment == "zy":
            kx, ky, kz = ky, kz, kx
        obj = PlaneWaveBasisByUnitVector(zip(kx, ky, kz, self.pol))
        obj.lattice = self.lattice
        obj.kpar = self.kpar
        return obj

    def kvecs(self, k0, material=Material(), modetype="up"):
        """Wave vectors.

        Args:
            k0 (float): Wave number.
            material (:class:`~treams.Material` or tuple, optional): Material
                definition. Defaults to vacuum/air.
            modetype (str, optional): Propagation direction. Defaults to "up".
        """
        if modetype not in ("up", "down"):
            raise ValueError("modetype not recognized")
        material = Material(material)
        kx = self._kx
        ky = self._ky
        kz = material.kzs(k0, kx, ky, self.pol) * (2 * (modetype == "up") - 1)
        if self.alignment == "yz":
            return kz, kx, ky
        if self.alignment == "zx":
            return ky, kz, kx
        return kx, ky, kz


def _raise_basis_error(*args):
    raise TypeError("'basis' must be BasisSet")


class PhysicsDict(util.AnnotationDict):
    """Physics dictionary.

    Derives from :class:`treams.util.AnnotationDict`. This dictionary has additionally
    several properties defined.

    Attributes:
        basis (:class:`BasisSet`): Basis set.
        k0 (float): Wave number.
        kpar (list): Parallel wave vector components. Usually, this is a list of length
            3 with its items corresponding to the Cartesian axes. Unspecified items are
            set to `nan`.
        lattice (:class:`~treams.Lattice`): Lattice definition.
        material (:class:`~treams.Material`): Material definition.
        modetype (str): Mode type, for spherical and cylindrical waves this can be
            "incident" and "scattered", for partial plane waves it can be "up" or
            "down".
        poltype (str): Polarization, see also :ref:`params:Polarizations`.
    """

    properties = {
        "basis": (
            ":class:`~treams.BasisSet`.",
            lambda x: isinstance(x, BasisSet),
            _raise_basis_error,
        ),
        "k0": ("Wave number.", lambda x: isinstance(x, float), float),
        "kpar": (
            "Wave vector components tangential to the lattice.",
            lambda x: isinstance(x, WaveVector),
            WaveVector,
        ),
        "lattice": (
            ":class:`~treams.Lattice`.",
            lambda x: isinstance(x, Lattice),
            Lattice,
        ),
        "material": (
            ":class:`~treams.Material`.",
            lambda x: isinstance(x, Material),
            Material,
        ),
        "modetype": ("Mode type.", lambda x: isinstance(x, str), str),
        "poltype": (":ref:`params:Polarizations`.", lambda x: isinstance(x, str), str,),
    }
    """Special properties tracked by the PhysicsDict."""

    def __setitem__(self, key, val):
        """Set item specified by key to the defined value.

        When overwriting an existing key an :class:`AnnotationWarning` is emitted.
        Avoid the warning by explicitly deleting the key first. The special attributes
        are cast to their corresponding types.

        Args:
            key (hashable): Key
            val : Value

        Warns:
            AnnotationWarning
        """
        if key not in self.properties:
            raise AttributeError(f"invalid key '{key}'")
        _, testfunc, castfunc = self.properties[key]
        if not testfunc(val):
            val = castfunc(val)
        super().__setitem__(key, val)


class PhysicsArray(util.AnnotatedArray):
    """Physics-aware array.

    A physics aware array is a special type of :class`~treams.util.AnnotatedArray`.
    Additionally to keeping track of the annotations, it is enhanced by the ability to
    create suiting linear operators to perform tasks like rotations, translations, or
    expansions into different basis sets and by applying special rules for the
    physical properties of :class:`PhysicsDict` upon matrix multiplications, see also
    :meth:`__matmul__`.
    """

    _scales = {"basis"}

    changepoltype = op.OperatorAttribute(op.ChangePoltype)
    """Polarization change matrix, see also :class:`treams.operators.ChangePoltype`."""
    efield = op.OperatorAttribute(op.EField)
    hfield = op.OperatorAttribute(op.HField)
    dfield = op.OperatorAttribute(op.DField)
    bfield = op.OperatorAttribute(op.BField)
    gfield = op.OperatorAttribute(op.GField)
    ffield = op.OperatorAttribute(op.FField)
    """Field evaluation matrix, see also :class:`treams.operators.EField`."""
    expand = op.OperatorAttribute(op.Expand)
    """Expansion matrix, see also :class:`treams.operators.Expand`."""
    expandlattice = op.OperatorAttribute(op.ExpandLattice)
    """Lattice expansion matrix, see also :class:`treams.operators.ExpandLattice`."""
    permute = op.OperatorAttribute(op.Permute)
    """Permutation matrix, see also :class:`treams.operators.Permute`."""
    rotate = op.OperatorAttribute(op.Rotate)
    """Rotation matrix, see also :class:`treams.operators.Rotate`."""
    translate = op.OperatorAttribute(op.Translate)
    """Translation matrix, see also :class:`treams.operators.Translate`."""

    def __init__(self, arr, ann=(), /, **kwargs):
        """Initialization."""
        super().__init__(arr, ann, **kwargs)
        self._check()

    @property
    def ann(self):
        """Array annotations."""
        # necessary to define the setter below
        return super().ann

    def __getattr__(self, key):
        try:
            return super().__getattr__(key)
        except AttributeError as err:
            if key in PhysicsDict.properties:
                return None
            raise err from None

    def __setattr__(self, key, val):
        if key in PhysicsDict.properties:
            val = (val,) * self.ndim if not isinstance(val, tuple) else val
            self.ann.as_dict[key] = val
        else:
            super().__setattr__(key, val)

    @ann.setter
    def ann(self, ann):
        """Set array annotations.

        See also :meth:`treams.util.AnnotatedArray.__setitem__`.
        """
        self._ann = util.AnnotationSequence(*(({},) * self.ndim), mapping=PhysicsDict)
        self._ann.update(ann)

    def __repr__(self):
        """String representation.

        For a more managable output only the special physics properties are shown
        alongside the array itself.
        """
        repr_arr = "    " + repr(self._array)[6:-1].replace("\n  ", "\n") + ","
        for key in PhysicsDict.properties:
            if getattr(self, key) is not None:
                repr_arr += f"\n    {key}={repr(getattr(self, key))},"
        return f"{self.__class__.__name__}(\n{repr_arr}\n)"

    def _check(self):
        """Run checks to validate the physical properties.

        The checks are run on the last two dimensions. They include:

            * Dispersion relation checks for :class:`PlaneWaveBasis` if `basis`, `k0`
              and `material` is defined`
            * `parity` polarizations are not physically permitted in chiral media.
            * All lattices explicitly given and in the basis hints must be compatible.
            * All tangential wave vector compontents must be compatible.
        """
        for a in self.ann[-2:]:
            k0 = a.get("k0")
            material = a.get("material")
            modetype = a.get("modetype")
            poltype = a.get("poltype")
            basis = a.get("basis", namedtuple("_basis", "lattice kpar")(None, None))
            if poltype == "parity" and getattr(material, "ischiral", False):
                raise ValueError("poltype 'parity' not permitted for chiral material")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Implement ufunc API.

        Additionally to keeping track of the annotations the special properties of a
        PhysicsArray are also "transparent" in matrix multiplications.
        """
        res = super().__array_ufunc__(ufunc, method, *inputs, **kwargs)
        if (
            ufunc is np.matmul
            and method == "__call__"
            and not isinstance(res, np.generic)
        ):
            axes = kwargs.get(
                "axes", [tuple(range(-min(np.ndim(i), 2), 0)) for i in inputs + (res,)]
            )
            anns = [
                i.ann[ax] if hasattr(i, "ann") else [{}, {}]
                for i, ax in zip(inputs, axes)
            ]
            for name in PhysicsDict.properties:
                if name in anns[0][-1] and all(name not in a for a in anns[1]):
                    res.ann[axes[-1][-1]].setdefault(name, anns[0][-1][name])
                if name in anns[1][0] and all(name not in a for a in anns[0]):
                    res.ann[axes[-1][0]].setdefault(name, anns[1][0][name])
        return res
